from __future__ import annotations
from pathlib import Path

import papermill as pm
import pytest


def notebooks():
    base = Path(__file__).parent.parent / "notebooks"
    return list(_ for _ in base.glob("**/*.ipynb") if "checkpoints" not in str(_))


# @pytest.skip(
#     "The tutorials use hard-coded paths!! cannot test them", allow_module_level=True
# )
@pytest.mark.parametrize("notebook", notebooks())
def test_notebook(notebook):
    base = Path(__file__).parent.parent / "notebooks"
    dst = notebook.relative_to(base)
    dst = Path(__file__).parent.parent / "tmp_notebooks" / str(dst)
    dst.parent.mkdir(exist_ok=True, parents=True)
    pm.execute_notebook(notebook, dst)


if __name__ == "__main__":
    print(notebooks())



__all__ = ["make_dinners_scenario"]

import itertools
import random
from typing import Iterable
import numpy as np
from anl2025.scenario import MultidealScenario
from anl2025.ufun import LambdaCenterUFun
from negmas import LinearUtilityAggregationFunction, make_issue, make_os

__all__ = ["make_dinners_scenario"]


class DinnersEvaluator:
    """Evaluates the center utility value of a set of agreements/disagreements"""

    def __init__(
        self,
        n_days: int,
        n_friends: int,
        reserved_value=0.0,
        values: dict[tuple[int, ...], float] | None = None,
    ):
        self.days = list(range(n_days))
        if values is None:
            # days = [self.days for _ in range(n_friends)]
            days = [[0, 1]] * n_days
            all_days = list(itertools.product(*days))
            v = np.random.rand(len(all_days))
            mn, mx = np.min(v), np.max(v)
            if np.all(np.abs(mx - mn)) > 1e-6:
                v -= mn
                v /= mx - mn
            else:
                v[:] = 1.0
            v[np.isnan(v)] = reserved_value
            values = dict(zip(all_days, v.tolist()))

        self.reserved_value = reserved_value
        self.n_days = len(self.days)
        self.values = values

    def __call__(self, agreements):
        if not agreements:
            return self.reserved_value
        outings = dict(zip(self.days, itertools.repeat(0)))
        for agreement in agreements:
            if agreement is None:
                continue
            # day is a tuple of one value which is the day selected
            outings[agreement[0]] += 1
        x = tuple(outings[day] for day in self.days)
        return self.values.get(x, self.reserved_value)


def make_dinners_scenario(
    n_friends: int = 3,
    n_days: int | None = None,
    friend_names: tuple[str, ...] | None = None,
    center_reserved_value: tuple[float, float] | float = 0.0,
    edge_reserved_values: tuple[float, float] = (0.0, 0.5),
    values: dict[tuple[int, ...], float] | None = None,
    public_graph: bool = True,
) -> MultidealScenario:
    """Creates a variation of the Dinners multideal scenario

    Args:
        n_friends: Number of friends.
        n_days: Number of days. If `None`, it will be the same as `n_friends`
        friend_names: Optionally, list of friend names (otherwise we will use Friend{i})
        center_reserved_value: The reserved value of the center. Either a number of a min-max range
        edge_reserved_valus: The reserved value of the friends (edges). Always, a min-max range
        values: A mapping from the number of dinners per day (a tuple of n_days integers) to utility value of the center
        public_graph: Should edges know n_edges and outcome_spaces?

    Returns:
        An initialized `MultidealScenario`.
    """
    if n_days is None:
        n_days = n_friends
    if not friend_names:
        friend_names = tuple(f"Friend{i+1}" for i in range(n_friends))
    assert (
        len(friend_names) == n_friends
    ), f"You passed {len(friend_names)} friend names but {n_friends=}"
    outcome_spaces = [
        make_os([make_issue(n_days, name="Day")], name=f"{name}Day")
        for name in friend_names
    ]
    if not isinstance(center_reserved_value, Iterable):
        r = float(center_reserved_value)
    else:
        r = center_reserved_value[0] + random.random() * (
            center_reserved_value[-1] - center_reserved_value[0]
        )
    return MultidealScenario(
        name="dinners",
        edge_ufuns=tuple(
            LinearUtilityAggregationFunction.random(
                os, reserved_value=edge_reserved_values, normalized=True
            )
            for os in outcome_spaces
        ),
        center_ufun=LambdaCenterUFun(
            outcome_spaces=outcome_spaces,
            evaluator=DinnersEvaluator(
                reserved_value=r,
                n_days=n_days,
                n_friends=n_friends,
                values=values,
            ),
            reserved_value=r,
        ),
        public_graph=public_graph,
    )



from pathlib import Path
from random import choice

from anl2025.scenario import MultidealScenario
from .dinners import *  # noqa: F403
from .job_hunt import *  # noqa: F403
from .target_quantity import *  # noqa: F403


__all__ = (
    dinners.__all__
    + job_hunt.__all__
    + target_quantity.__all__
    + ["load_example_scenario", "get_example_scenario_names"]
)  # type: ignore # noqa: F405


def _base_path():
    path = Path(__file__).parent
    if path.name == "scenarios":
        path = path.parent
    return path / "example_scenarios"


def load_example_scenario(name: str | None = None) -> MultidealScenario:
    """Loads an example scenario.

    Remarks:
        - Currently the following scenarios are available: Dinners and TargetQuantity.
        - If you do not pass any name a randomly chosen example scenario will be returned.
    """
    if not name:
        paths = [_.resolve() for _ in _base_path().glob("*") if _.is_dir()]
    else:
        paths = [_base_path() / name]
    s = MultidealScenario.from_folder(choice(paths))
    assert s is not None
    return s


def get_example_scenario_names() -> list[str]:
    """Returns a list of example scenario names."""
    return [_.name for _ in _base_path().glob("*") if _.is_dir()]



from pathlib import Path

from anl2025.runner import MultidealScenario
from .common import TYPE_IDENTIFIER

__all__ = ["load_multideal_scenario"]


def load_multideal_scenario(
    folder: Path | str,
    name: str | None = None,
    edges_know_details: bool = True,
    python_class_identifier: str = TYPE_IDENTIFIER,
    type_marker=f"{TYPE_IDENTIFIER}:",
) -> MultidealScenario | None:
    """
    Loads a multi-deal scenario from the given folder.

    Args:
        folder: The path to load the scenario from
        name: The name to give to the scenario. If not given, the folder name
        edges_know_details: If given, edge ufuns will have `n_edges`, `outcome_spaces` members
                            that reveal the number of edges in total and the outcome space for each
                            negotiation thread.
        python_class_identifier: the key in the yaml to define a type.
        type_marker: A marker at the beginning of a string to define a type (for future proofing).
    """
    folder = folder if isinstance(folder, Path) else Path(folder)
    return MultidealScenario.from_folder(
        folder,
        name=name,
        public_graph=edges_know_details,
        python_class_identifier=python_class_identifier,
        type_marker=type_marker,
    )



from copy import deepcopy
from time import perf_counter
from rich import print
from typing import Any
from random import choice
import pandas as pd
from attr import define, field
from pathlib import Path
import matplotlib.pyplot as plt
from negmas import ControlledNegotiator
from negmas.outcomes import Outcome
from negmas.sao import SAOMechanism
from negmas.helpers import unique_name

from anl2025.ufun import CenterUFun, make_side_ufun
from anl2025.negotiator import (
    ANL2025Negotiator,
    Boulware2025,
    Conceder2025,
    Linear2025,
    Random2025,
)
from anl2025.scenario import MultidealScenario, make_multideal_scenario
from anl2025.common import SEQUENTIAL_METHOD, get_agent_class, RunParams, DEFAULT_METHOD


__all__ = [
    "run_session",
    "run_generated_session",
    "SessionResults",
    "RunParams",
    "AssignedScenario",
    "assign_scenario",
]

TRACE_COLS = (
    "time",
    "relative_time",
    "step",
    "negotiator",
    "offer",
    "responses",
    "state",
)


@define
class SessionResults:
    """Results of a single multideal negotiation

    Attributes:
        mechanisms: The mechanisms representing negotiation threads.
        center: The center agent
        edges: Edge agents
        agreements:  Negotiation outcomes for all threads.
        center_utility: The utility received by the center.
        edge_utilities: The utilities of all edges.
    """

    mechanisms: list[SAOMechanism]
    center: ANL2025Negotiator
    edges: list[ANL2025Negotiator]
    agreements: list[Outcome | None]
    center_utility: float
    edge_utilities: list[float]
    n_succeeded: int = field(init=False)
    n_timedout: int = field(init=False)
    n_failed: int = field(init=False)
    total_time: float
    times: list[float]

    def __attrs_post_init__(self):
        self.n_succeeded = len([_ for _ in self.agreements if _ is not None])
        self.n_timedout = len([_ for _ in self.mechanisms if _.state.timedout])
        self.n_failed = len(self.agreements) - self.n_succeeded - self.n_timedout

    # final_states: list[SAOState]


@define
class AssignedScenario:
    """A scenario with assigned agents ready to run.

    Attributes:
        scenario: The scenario
        run_params: `RunParams` controlling how the session is run
        center: The center agent
        edges: edge agents
    """

    scenario: MultidealScenario
    run_params: RunParams
    center: ANL2025Negotiator
    edges: list[ANL2025Negotiator]
    _saved_scenario: MultidealScenario = field(init=False)

    def __attrs_post_init__(self):
        self._saved_scenario = deepcopy(self.scenario)

    def run(
        self,
        name: str = "",
        output: Path | str | None = None,
        verbose: bool = False,
        dry: bool = False,
    ) -> SessionResults:
        """Runs a multi-deal negotiation and gets the results"""
        if output and isinstance(output, str):
            output = Path(output)

        def type_name(x):
            if isinstance(x, Boulware2025):
                return "Boulware2025"
            if not issubclass(x.default_negotiator_type, ControlledNegotiator):
                return f"ANL2025({x.default_negotiator_type.__name__})"
            return x.__class__.__name__.split(".")[-1]

        center = self.center
        edges = self.edges
        # edge_ufuns = [deepcopy(_) for _ in self.scenario.edge_ufuns]
        edge_ufuns = self.scenario.edge_ufuns
        center_ufun = self.scenario.center_ufun
        # center_ufun = deepcopy(self.scenario.center_ufun)
        # We MUST reconnect side ufuns for expected outcomes to work
        # center_ufun._effective_side_ufuns = tuple(
        #     make_side_ufun(center_ufun, i, None) for i in range(center_ufun.n_edges)
        # )
        # assume stationarity
        center_ufun.stationary = True
        center_ufun.stationary_sides = (
            center_ufun.stationary_sides or self.run_params.method == SEQUENTIAL_METHOD
        )
        len(edge_ufuns)
        if verbose:
            print(f"Adding center of type {type_name(center)}")

        mechanisms = []
        try:
            side_ufuns = center_ufun.side_ufuns()
        except Exception:
            side_ufuns = None
        if not side_ufuns:
            side_ufuns = [None] * len(edges)
        side_ufuns = [
            make_side_ufun(center_ufun, i, side) for i, side in enumerate(side_ufuns)
        ]
        for i, (edge_ufun, side_ufun, edge) in enumerate(
            zip(edge_ufuns, side_ufuns, edges, strict=True)
        ):
            params_ = {k: v for k, v in self.run_params.mechanism_params.items()}
            params_ |= dict(
                name=f"n{i}",
                n_steps=self.run_params.nsteps,
                end_on_no_response=self.run_params.end_on_no_response,
                time_limit=self.run_params.time_limit,
                hidden_time_limit=self.run_params.hidden_time_limit,
                outcome_space=edge_ufun.outcome_space,
                one_offer_per_step=self.run_params.atomic,
            )
            if self.run_params.ignore_negotiator_exceptions is not None:
                params_[
                    "ignore_negotiator_exceptions"
                ] = self.run_params.ignore_negotiator_exceptions
            m = SAOMechanism(**params_)
            m.id = m.name = f"n{i}"
            if verbose:
                print(f"Adding edge {i} of type {type_name(edge)} (thread: {m.name})")
            m.add(
                center.create_negotiator(
                    cntxt=dict(center=True, ufun=side_ufun, index=i),
                    ufun=side_ufun,
                    id=f"s{i}",
                    private_info=dict(opponent_ufun=edge_ufun)
                    if self.run_params.share_ufuns
                    else dict(),
                )
            )
            m.negotiators[-1].id = m.negotiators[-1].name = f"s{i}"
            m.add(
                edge.create_negotiator(
                    cntxt=dict(center=False, ufun=edge_ufun, index=i),
                    ufun=edge_ufun,
                    id=f"e{i}",
                    private_info=dict(opponent_ufun=side_ufun)
                    if self.run_params.share_ufuns
                    else dict(),
                )
            )
            m.negotiators[-1].id = m.negotiators[-1].name = f"e{i}"
            mechanisms.append(m)
        assert isinstance(center.ufun, CenterUFun)
        center.init()
        for edge in edges:
            edge.init()
        if dry:
            return SessionResults(
                mechanisms=mechanisms,
                center=center,
                center_utility=0.0,
                edge_utilities=[0.0] * len(edges),
                edges=edges,
                agreements=[None] * len(edges),
                total_time=0.0,
                times=[0.0] * len(mechanisms),
                # final_states=[_.state for _ in mechanisms],
            )

        base = None
        if output:
            base = output / name
            (base / "log").mkdir(parents=True, exist_ok=True)
            (base / "plots").mkdir(parents=True, exist_ok=True)

        def plot_result(i, m, base=base):
            assert base is not None
            df = pd.DataFrame(data=m.full_trace, columns=TRACE_COLS)  # type: ignore
            df.to_csv(base / "log" / f"{m.id}.csv", index_label="index")
            m.plot(save_fig=True, path=str(base / "plots"), fig_name=f"n{i}.png")
            plt.close()

        _strt = perf_counter()
        # be compatible with negmas before 0.11.4
        try:
            SAOMechanism.runall(
                mechanisms,
                method=self.run_params.method,
                keep_order=self.run_params.keep_order,
                completion_callback=plot_result if output else None,
                ignore_mechanism_exceptions=self.run_params.ignore_mechanism_exceptions,  # type: ignore
            )  # type: ignore
        except Exception:
            SAOMechanism.runall(
                mechanisms,
                method=self.run_params.method,
                keep_order=self.run_params.keep_order,
                completion_callback=plot_result if output else None,
            )  # type: ignore
        total_time = perf_counter() - _strt
        if not name:
            name = unique_name("session", sep=".")
        agreements = [_.agreement for _ in mechanisms]

        return SessionResults(
            mechanisms=mechanisms,
            center=center,
            agreements=agreements,
            center_utility=float(self._saved_scenario.center_ufun(tuple(agreements))),
            edge_utilities=[
                float(edge_ufun(_)) if edge_ufun else float("nan")
                for edge_ufun, _ in zip(self._saved_scenario.edge_ufuns, agreements)
            ],
            edges=edges,
            total_time=total_time,
            times=[m.time for m in mechanisms],
            # final_states=[_.state for _ in mechanisms],
        )


def assign_scenario(
    scenario: MultidealScenario,
    run_params: RunParams,
    center_type: str | type[ANL2025Negotiator] = "Boulware2025",
    center_params: dict[str, Any] | None = None,
    edge_types: list[str | type[ANL2025Negotiator]] = [
        Boulware2025,
        Random2025,
        Linear2025,
        Conceder2025,
    ],
    edge_params: list[dict[str, Any]] | None = None,
    verbose: bool = False,
    sample_edges: bool = False,
) -> AssignedScenario:
    """Assigns a multidal scenario to negotiators

    Args:
        scenario: The multideal scenario
        run_params: Parameters controlling the session is run (See `RunParams`).
        center_type: Type of the center agent
        center_params: Optional parameters to pass to the center agent when it is being constructed
        edge_types: Types of edge agents
        edge_params: Optional parameters to pass to edge agents.
        verbose: Print progress
        sample_edges: If true, the `edge_types` will be used as a pool to sample from
                      instead of being assigned to edges in order

    Returns:
        An `AssignedScenario` ready to run.
    """
    center_ufun = scenario.center_ufun
    edge_ufuns = scenario.edge_ufuns
    nedges = len(edge_ufuns)

    if not edge_params:
        edge_params = [dict() for _ in range(nedges)]
    center_params = center_params if center_params else dict()
    center = get_agent_class(center_type)(
        id="center", ufun=center_ufun, **center_params
    )

    agents = [get_agent_class(_) for _ in edge_types]
    edges: list[ANL2025Negotiator] = []
    if verbose:
        print(
            f"Center type={center.__class__.__name__}\nWill use the following agents for edges\n{[_.__name__ if not isinstance(_, str) else _.split('.')[-1] for _ in agents]}"
        )
    for i, (edge_ufun, edge_p) in enumerate(zip(edge_ufuns, edge_params)):
        if sample_edges:
            edget = choice(agents)
        else:
            edget = agents[i % len(edge_types)]
        edge = edget(ufun=edge_ufun, id=f"edge{i}", n_edges=nedges, **edge_p)
        edges.append(edge)
    assert isinstance(center.ufun, CenterUFun)
    return AssignedScenario(
        scenario=scenario,
        run_params=run_params,
        center=center,
        edges=edges,
    )


def run_generated_session(
    # center
    center_type: str = "Boulware2025",
    center_params: dict[str, Any] | None = None,
    center_reserved_value_min: float = 0.0,
    center_reserved_value_max: float = 0.0,
    center_ufun_type: str | type[CenterUFun] = "MaxCenterUFun",
    center_ufun_params: dict[str, Any] | None = None,
    # edges
    nedges: int = 10,
    edge_reserved_value_min: float = 0.1,
    edge_reserved_value_max: float = 0.4,
    edge_types: list[str | type[ANL2025Negotiator]] = [
        Boulware2025,
        Random2025,
        Linear2025,
        Conceder2025,
    ],
    # outcome space
    nissues: int = 3,
    nvalues: int = 7,
    # mechanism params
    nsteps: int = 100,
    keep_order: bool = True,
    share_ufuns: bool = True,
    atomic: bool = False,
    # output and logging
    output: Path | str | None = Path.home() / "negmas" / "anl2025" / "session",
    name: str = "",
    dry: bool = False,
    method=DEFAULT_METHOD,
    verbose: bool = False,
) -> SessionResults:
    """Generates a multideal negotiation session and runs it.

    Args:
        method: the method to use for running all the sessions.
                Acceptable options are: sequential, ordered, threads, processes.
                See `negmas.mechanisms.Mechanism.run_all()` for full details.
        center_type: Type of the center agent
        center_params: Optional parameters to pass to the center agent.
        center_reserved_value_min: Minimum reserved value for the center agent.
        center_reserved_value_max: Maximum reserved value for the center agent.
        center_ufun_type: Type of the center agent ufun.
        center_ufun_params: Parameters to pass to the center agent ufun.
        nedges: Number of edges.
        edge_reserved_value_min: Minimum reserved value for edges.
        edge_reserved_value_max: Maximum reserved value for edges.
        edge_types: Types of edge agents
        nissues: Number of issues to use for each thread.
        nvalues: Number of values per issue for each thread.
        nsteps: Number of negotiation steps.
        keep_order: Keep the order of edges when advancing the negotiation.
        share_ufuns: If given, agents will have access to partner ufuns through `self.opponent_ufun`.
        atomic: If given, one step corresponds to one offer instead of a full round.
        output: Folder to store the logs and results within.
        name: Name of the session
        dry: IF true, nothing will be run.
        verbose: Print progress

    Returns:
        `SessionResults` giving the results of the multideal negotiation session.
    """
    if output and isinstance(output, str):
        output = Path(output)
    sample_edges = nedges > 0
    if not sample_edges:
        nedges = len(edge_types)
    scenario = make_multideal_scenario(
        nedges=nedges,
        nissues=nissues,
        nvalues=nvalues,
        center_reserved_value_min=center_reserved_value_min,
        center_reserved_value_max=center_reserved_value_max,
        center_ufun_type=center_ufun_type,
        center_ufun_params=center_ufun_params,
        edge_reserved_value_min=edge_reserved_value_min,
        edge_reserved_value_max=edge_reserved_value_max,
    )
    run_params = RunParams(
        nsteps=nsteps,
        keep_order=keep_order,
        share_ufuns=share_ufuns,
        atomic=atomic,
        method=method,
    )
    assigned = assign_scenario(
        scenario=scenario,
        run_params=run_params,
        center_type=center_type,
        center_params=center_params,
        edge_types=edge_types,
        verbose=verbose,
        sample_edges=sample_edges,
    )
    return assigned.run(output=output, name=name, dry=dry, verbose=verbose)


def run_session(
    scenario: MultidealScenario,
    # center
    center_type: str | type[ANL2025Negotiator] = "Boulware2025",
    center_params: dict[str, Any] | None = None,
    # edges
    edge_types: list[str | type[ANL2025Negotiator]] = [
        Boulware2025,
        Linear2025,
        Conceder2025,
    ],
    # mechanism params
    nsteps: int = 100,
    keep_order: bool = True,
    share_ufuns: bool = True,
    atomic: bool = False,
    # output and logging
    output: Path | None = Path.home() / "negmas" / "anl2025" / "session",
    name: str = "",
    dry: bool = False,
    method=DEFAULT_METHOD,
    verbose: bool = False,
    sample_edges: bool = False,
) -> SessionResults:
    """Runs a multideal negotiation session and runs it.

    Args:
        scenario: The negotiation scenario (must be `MultidealScenario`).
        method: the method to use for running all the sessions.
                Acceptable options are: sequential, ordered, threads, processes.
                See `negmas.mechanisms.Mechanism.run_all()` for full details.
        center_type: Type of the center agent
        center_params: Optional parameters to pass to the center agent.
        center_ufun_type: Type of the center agent ufun.
        center_ufun_params: Parameters to pass to the center agent ufun.
        edge_types: Types of edge agents
        nsteps: Number of negotiation steps.
        keep_order: Keep the order of edges when advancing the negotiation.
        share_ufuns: If given, agents will have access to partner ufuns through `self.opponent_ufun`.
        atomic: If given, one step corresponds to one offer instead of a full round.
        output: Folder to store the logs and results within.
        name: Name of the session
        dry: IF true, nothing will be run.
        verbose: Print progress
        sample_edges: If true, the `edge_types` will be used as a pool to sample from otherwise edges will
                      be of the types defined by edge_types in order

    Returns:
        `SessionResults` giving the results of the multideal negotiation session.
    """
    if output and isinstance(output, str):
        output = Path(output)
    run_params = RunParams(
        nsteps=nsteps,
        keep_order=keep_order,
        share_ufuns=share_ufuns,
        atomic=atomic,
        method=method,
    )
    # if not sample_edges:
    #     assert (
    #         len(edge_types) == len(scenario.edge_ufuns)
    #     ), f"You are trying to run a session without sampling, but the number of provided edge types ({len(edge_types)}) is not equal to the number of edge ufuns in the scenario ({len(scenario.edge_ufuns)})."

    assigned = assign_scenario(
        scenario=scenario,
        run_params=run_params,
        center_type=center_type,
        center_params=center_params,
        edge_types=edge_types,
        verbose=verbose,
        sample_edges=sample_edges,
    )
    return assigned.run(output=output, name=name, dry=dry, verbose=verbose)



from collections import defaultdict
from attr import asdict, field
from multiprocessing import cpu_count
from concurrent.futures import ProcessPoolExecutor, as_completed
from negmas.helpers import unique_name
from negmas.serialization import dump
from rich import print
from rich.progress import track
from collections.abc import Sequence
from typing import TypedDict
from pathlib import Path
from typing import Self
import random
from anl2025.ufun import CenterUFun
from negmas.helpers.types import get_class, get_full_type_name
from negmas.serialization import serialize, deserialize
from negmas.helpers.inout import load
from typing import Any
from anl2025.negotiator import ANL2025Negotiator
from anl2025.negotiator import (
    TimeBased2025,
    Random2025,
    Boulware2025,
    Linear2025,
    Conceder2025,
)
from anl2025.runner import (
    AssignedScenario,
    MultidealScenario,
    RunParams,
    SessionResults,
    assign_scenario,
    make_multideal_scenario,
)
from anl2025.common import DEFAULT_METHOD, TYPE_IDENTIFIER
from attr import define

__all__ = [
    "Tournament",
    "TournamentResults",
    "anl2025_tournament",
    "DEFAULT_TOURNAMENT_PATH",
    "DEFAULT_ANL2025_COMPETITORS",
]

DEFAULT_TOURNAMENT_PATH = Path.home() / "negmas" / "anl2025" / "tournaments"
"""Default location to store tournament logs"""

DEFAULT_ANL2025_COMPETITORS = (
    TimeBased2025,
    Random2025,
    Boulware2025,
    Linear2025,
    Conceder2025,
    # IndependentBoulware2025,
    # IndependentLinear2025,
    # IndependentConceder2025,
)


class ScoreRecord(TypedDict):
    """Score of a single run for a single agent

    Attributes:
        agent: The agent being scored
        utility: Utility of the agent
        partner_average_utility: Average utilities of agent partners
        scenario: The scenario on which this score was received.
        repetition: The repetition number of this run.
        rotation: The rotation number of this run.
        scenario_index: Index of the scenario.
        index: Index of the agent. center = 0 and edges start at 1
        errors: Number of errors of this agent
        partner_errors: Number of errors of the opponent (partner)
        mechanism_errors: Number of mechanism errors
    """

    agent: str
    utility: float
    partner_average_utility: float
    scenario: str
    repetition: int
    rotation: int
    scenario_index: int
    index: int
    errors: int
    partner_errors: int
    mechanism_errors: int
    time: float


@define
class JobInfo:
    assigned: AssignedScenario
    output: Path | None
    sname: str
    rep_index: int
    competitor_index: int
    scenario_index: int
    center: type
    center_params: dict[str, Any] | None
    edges: tuple[type, ...] | list[type]
    edge_params: tuple[dict, ...] | list[dict]
    edge_info: list[tuple[type, dict[str, Any] | None]]
    nedges_counted: int


@define
class SessionInfo:
    """Information of a single negotiation during a tournament"""

    scenario_name: str
    repetition: int
    rotation: int
    center_type_name: str
    center_params: dict[str, Any]
    edge_type_names: list[str]
    edge_params: list[dict[str, Any] | None] | tuple[dict[str, Any] | None, ...]
    results: SessionResults
    path: Path | None = None


@define
class TournamentResults:
    """Results of a tournament"""

    final_scores: dict[str, float]
    final_scoresE: dict[str, float]
    final_scoresC: dict[str, float]
    center_count: dict[str, float]
    edge_count: dict[str, float]
    weighted_average: dict[str, float]
    scores: list[ScoreRecord]
    session_results: list[SessionInfo]
    path: Path | None = None
    n_threads_succeeded: int = field(init=False)
    n_threads_timedout: int = field(init=False)
    n_threads_failed: int = field(init=False)

    def __attrs_post_init__(self):
        self.n_threads_succeeded = sum(
            [_.results.n_succeeded for _ in self.session_results]
        )
        self.n_threads_timedout = sum(
            [_.results.n_timedout for _ in self.session_results]
        )
        self.n_threads_failed = sum([_.results.n_failed for _ in self.session_results])


def run_session(job: JobInfo, dry: bool, verbose: bool) -> tuple[JobInfo, SessionInfo]:
    if verbose:
        print(f"Scenario {job.assigned.scenario.name}")
    assigned = job.assigned
    output = job.output
    sname = job.sname
    i = job.rep_index
    j = job.competitor_index
    center = job.center
    center_params = job.center_params
    edges = job.edges
    edge_params = job.edge_params
    r = assigned.run(
        output=output,
        name=f"{sname}_{j}_{i}",
        dry=dry,
        verbose=verbose,
    )
    return job, SessionInfo(
        scenario_name=sname,
        repetition=i,
        rotation=j,
        center_type_name=get_full_type_name(center),
        center_params=center_params if center_params else dict(),
        edge_type_names=[get_full_type_name(_) for _ in edges],
        edge_params=edge_params,  # type: ignore
        results=r,
    )


def anl2025_tournament(
    scenarios: tuple[MultidealScenario, ...],
    competitors: tuple[str | type[ANL2025Negotiator], ...],
    n_repetitions: int = 3,
    n_steps: int = 100,
    competitor_params: tuple[dict[str, Any] | None, ...] | None = None,
    path: Path | str | None = None,
    no_double_scores: bool = True,
    non_comptitor_types: tuple[str | type[ANL2025Negotiator], ...] | None = None,
    non_comptitor_params: tuple[dict[str, Any], ...] | None = None,
    n_jobs: int | float | None = 0,
    center_multiplier: float | None = None,
    edge_multiplier: float = 1,
    verbose: bool = False,
    dry: bool = False,
    keep_order: bool = True,
    share_ufuns: bool = False,
    atomic: bool = False,
    method: str = DEFAULT_METHOD,
) -> TournamentResults:
    """Creates and runs a tournament.

    Args:
        scenarios: The scenarios to use for the tournament
        competitors: The competitor negotiators
        n_repetitions: Number of repetitions of each configuration of agents
        n_steps: Number of steps of each negotiation
        competitor_params: Optional parameters to use for constructing the competitors
        path: Path to store logs and results of the tournament.
        no_double_scores: Avoid having the same agent in multiple positions in the same negotiation
        non_comptitor_types: Types to use to fill missing edge locations if not enough competitors are available
        non_comptitor_params: Paramters of non-competitor-types
        n_jobs: Number of parallel jobs to use.
                None (and negative numbers) mean serially, 0 means use all cores, fractions mean fraction of available
                cores, integers mean exact number of cores
        center_multiplier: A number to multiply center utilities with before calculating the score. Can be used
                           to give more or less value to being a center. If None, it will be equal to the number of edges.
        edge_multiplier: A number to multiply edge utilities with before calculating the score. Can be used
                           to give more or less value to being an edge
        verbose: Print progress messages
        dry: If given, the tournament will be created but will not be run.
        keep_order: Keep the order of edges when running a session
        share_ufuns: Allow negotiators to access partner utility function as `self.opponent_ufun`
        atomic: If true, every step is on offer, otherwise, every step is a complete round (two offers)
        method: The method for stepping negotiation threads. All methods supported by `negmas.Mechanism.runall()`
                are supported including sequential which means completing one negotiation before starting the next.

    Returns:
        [TODO:return]
    """
    run_params = RunParams(
        nsteps=n_steps,
        keep_order=keep_order,
        share_ufuns=share_ufuns,
        atomic=atomic,
        method=method,
    )
    tournament = Tournament(
        competitors=competitors,
        scenarios=scenarios,
        run_params=run_params,
        competitor_params=competitor_params,
    )
    return tournament.run(
        path=path,
        n_repetitions=n_repetitions,
        verbose=verbose,
        dry=dry,
        no_double_scores=no_double_scores,
        non_comptitor_types=non_comptitor_types,
        non_comptitor_params=non_comptitor_params,
        n_jobs=n_jobs,
        center_multiplier=center_multiplier,
        edge_multiplier=edge_multiplier,
    )


def make_tournament_name():
    return unique_name("t", sep="_")


@define
class Tournament:
    """Represents a tournament

    Attributes:
        competitors: the competing agents of type `ANL2025Negotiator` each
        scenarios: the scenarios in which the competitors are tested
        run_params: parameters controlling the tournament run (See `RunParams`)
        competitor_params: Parameters to pass to the competitors
    """

    competitors: tuple[str | type[ANL2025Negotiator], ...]
    scenarios: tuple[MultidealScenario, ...]
    run_params: RunParams
    competitor_params: tuple[dict[str, Any] | None, ...] | None = None
    name: str = field(factory=make_tournament_name)

    @classmethod
    def from_scenarios(
        cls,
        competitors: Sequence[str | type[ANL2025Negotiator]],
        run_params: RunParams,
        scenarios: tuple[MultidealScenario, ...] = tuple(),
        n_generated: int = 0,
        nedges: int = 3,
        nissues: int = 3,
        nvalues: int = 7,
        # edge ufuns
        center_reserved_value_min: float = 0.0,
        center_reserved_value_max: float = 0.0,
        center_ufun_type: str | type[CenterUFun] = "MaxCenterUFun",
        center_ufun_params: dict[str, Any] | None = None,
        # edge ufuns
        edge_reserved_value_min: float = 0.1,
        edge_reserved_value_max: float = 0.4,
        competitor_params: tuple[dict[str, Any] | None, ...] | None = None,
        name: str | None = None,
    ) -> Self:
        """Loads a tournament from the given scenarios (optionally generating new ones)

        Args:
            competitors: Competing agents
            run_params: `RunParams` controlling the timing of each multideal negotiation
            scenarios: An optional tuple of predefined scenarios (`MultidealScenario`)
            n_generated: Number of new scenarios to generate
            nedges: Number of negotiation threads (only used if `n_generated` > 0)
            nissues:Number of negotiation issues per thread (only used if `n_generated` > 0)
            nvalues: Number of values per issue (only used if `n_generated` > 0)
            center_reserved_value_min: Minimum reserved value of the center for generated scenarios.
            center_reserved_value_max: Maximum reserved value of the center for generated scenarios.
            center_ufun_type: center agent ufun for generated scenarios.
            center_ufun_params: center agent ufun params for generated scenarios.
            edge_reserved_value_min: Minimum reserved value of  edges for generated scenarios.
            edge_reserved_value_max: Maximum reserved value of  edges for generated scenarios.
            competitor_params: Optional competitor paramters

        Returns:
            A `Tournament` ready to run
        """
        # if nedges > len(competitors):
        #     raise ValueError(
        #         f"We have {len(competitors)} competitors which is not enough for {nedges} edges"
        #     )
        return cls(
            name=name if name else make_tournament_name(),
            competitors=tuple(competitors),
            competitor_params=competitor_params,
            run_params=run_params,
            scenarios=tuple(
                list(scenarios)
                + [
                    make_multideal_scenario(
                        nedges=nedges,
                        nissues=nissues,
                        nvalues=nvalues,
                        center_reserved_value_min=center_reserved_value_min,
                        center_reserved_value_max=center_reserved_value_max,
                        center_ufun_type=center_ufun_type,
                        center_ufun_params=center_ufun_params,
                        edge_reserved_value_min=edge_reserved_value_min,
                        edge_reserved_value_max=edge_reserved_value_max,
                    )
                    for _ in range(n_generated)
                ]
            ),
        )

    def __attrs_post_init__(self):
        if not self.competitor_params:
            self.competitor_params = tuple(dict() for _ in range(len(self.competitors)))
        self.competitor_params = tuple(
            dict() if not _ else _ for _ in self.competitor_params
        )

    def save(
        self,
        path: Path | str,
        separate_scenarios: bool = False,
        python_class_identifier=TYPE_IDENTIFIER,
    ):
        """
        Saves the tournament information.

        Args:
            path: A file to save information about the tournament to
            separate_scenarios: If `True`, scenarios will be saved inside a `scenarios` folder beside the path given otherwise they will be included in the file
        """
        path = path if isinstance(path, Path) else Path(path)
        data = dict(
            competitors=[get_full_type_name(_) for _ in self.competitors],
            run_params=asdict(self.run_params),
            competitor_params=None
            if not self.competitor_params
            else [
                serialize(_, python_class_identifier=python_class_identifier)
                for _ in self.competitor_params
            ],
        )
        if separate_scenarios:
            base = path.resolve().parent / "scenarios"
            for i, s in enumerate(self.scenarios):
                name = s.name if s.name else f"s{i:03}"
                dst = base
                dst.mkdir(parents=True, exist_ok=True)
                dump(
                    serialize(s, python_class_identifier=python_class_identifier),
                    dst / f"{name}.yaml",
                )
        else:
            data["scenarios"] = [
                serialize(_, python_class_identifier=python_class_identifier)
                for _ in self.scenarios
            ]
        dump(data, path)

    @classmethod
    def load(cls, path: Path | str, python_class_identifier=TYPE_IDENTIFIER):
        """Loads the tournament information."""

        path = path if isinstance(path, Path) else Path(path)
        info = load(path)
        base = path.resolve().parent / "scenarios"
        if "scenarios" not in info:
            info["scenarios"] = []
        else:
            info["scenarios"] = list(info["scenarios"])

        if base.exists():
            info["scenarios"] += [
                deserialize(f, python_class_identifier=python_class_identifier)
                for f in base.glob("*.yaml")
            ]

        return cls(
            competitors=info["competitors"],
            scenarios=[
                deserialize(_, python_class_identifier=python_class_identifier)
                for _ in info["scenarios"]
            ],  # type: ignore
            run_params=RunParams(**info["run_params"]),
            competitor_params=None  # type: ignore
            if not info.get("competitor_params", None)
            else deserialize(
                info["competitor_params"],
                python_class_identifier=python_class_identifier,
            ),
        )

    def run(
        self,
        n_repetitions: int,
        path: Path | str | None = None,
        verbose: bool = False,
        dry: bool = False,
        no_double_scores: bool = True,
        non_comptitor_types: tuple[str | type[ANL2025Negotiator], ...] | None = None,
        non_comptitor_params: tuple[dict[str, Any], ...] | None = None,
        n_jobs: int | float | None = 0,
        center_multiplier: float | None = None,
        edge_multiplier: float = 1,
    ) -> TournamentResults:
        """Run the tournament

        Args:
            n_repetitions: Number of repetitions of rotations over scenarios
            path: Path to save the results to
            verbose: Print progress
            dry: Do not really run the negotiations.
            no_double_scores: Avoid having the same agent in multiple positions in the same negotiation
            non_comptitor_types: Types to use to fill missing edge locations if not enough competitors are available
            non_comptitor_params: Paramters of non-competitor-types
            n_jobs: Number of parallel jobs to use.
                    None (and negative numbers) mean serially, 0 means use all cores, fractions mean fraction of available
                    cores, integers mean exact number of cores
            center_multiplier: A number to multiply center utilities with before calculating the score. Can be used
                               to give more or less value to being a center. If None, it will be equal to the number of edges.
            edge_multiplier: A number to multiply edge utilities with before calculating the score. Can be used
                               to give more or less value to being an edge

        Returns:
            `TournamentResults` with all scores and final-scores
        """
        if path is not None:
            path = path if isinstance(path, Path) else Path(path)
        if n_jobs is not None:
            if isinstance(n_jobs, float) and n_jobs < 1.0:
                n_jobs = int(0.5 + cpu_count() * n_jobs)
            elif isinstance(n_jobs, float):
                n_jobs = int(0.5 + n_jobs)
            if n_jobs < 0:
                n_jobs = None
            elif n_jobs == 0:
                n_jobs = cpu_count()

        results = []
        assert isinstance(self.competitor_params, tuple)
        final_scores = defaultdict(float)
        final_scoresC = defaultdict(float)
        final_scoresE = defaultdict(float)
        count_edge = defaultdict(float)
        count_center = defaultdict(float)

        scores = []
        center_multiplier_val = center_multiplier

        def type_name(x):
            return get_full_type_name(x).replace("anl2025.negotiator.", "")

        if non_comptitor_types:
            non_comptitor_types = tuple(get_class(_) for _ in non_comptitor_types)
            non_comptitor_params = (
                non_comptitor_params
                if non_comptitor_params
                else tuple(dict() for _ in range(len(non_comptitor_types)))
            )
            non_competitors = [
                (n, p)
                for n, p in zip(non_comptitor_types, non_comptitor_params, strict=True)
            ]
        else:
            non_competitors = None

        jobs = []

        for i in track(range(n_repetitions), "Preparing Negotiation Sessions"):
            competitors = [
                (get_class(c), p)
                for c, p in zip(self.competitors, self.competitor_params, strict=True)
            ]
            for k, scenario in enumerate(self.scenarios):
                nedges = len(scenario.edge_ufuns)
                sname = scenario.name if scenario.name else f"s{k:03}"
                random.shuffle(competitors)
                # put each competitor in center once per scenario
                for j in range(len(competitors)):
                    if len(competitors) >= nedges + 1:
                        players = [_ for _ in competitors]
                        # players = competitors[: nedges + 1]
                    else:
                        # add extra players at the end if not enough competitors are available
                        players = competitors + list(
                            random.choices(
                                non_competitors if non_competitors else competitors,
                                k=nedges + 1 - len(competitors),
                            )
                        )
                    # ignore the randomly added edges if no-double-scores is set
                    nedges_counted = (
                        nedges
                        if not no_double_scores
                        else min(len(competitors) - 1, nedges)
                    )
                    if path:
                        output = path / "results" / sname / f"r{j:03}t{i:03}"
                    else:
                        output = None
                    # if verbose:
                    #     print(f"{j=}, {players=}")
                    center, center_params = players[j]
                    edge_info = [_ for _ in players[:j] + players[j + 1 :]][
                        : nedges + 1
                    ]
                    # not sure if the following shuffle is useful!
                    # It tries to randomize the order of the edges to avoid
                    # having a systematic bias but we randomize competitors anyway.
                    random.shuffle(edge_info)
                    edges = [_[0] for _ in edge_info]
                    edge_params = [_[1] if _[1] else dict() for _ in edge_info]
                    assigned = assign_scenario(
                        scenario=scenario,
                        run_params=self.run_params,
                        center_type=center,
                        center_params=center_params,
                        edge_types=edges,  # type: ignore
                        edge_params=edge_params,
                        verbose=verbose,
                        sample_edges=False,
                    )
                    jobs.append(
                        JobInfo(
                            assigned,
                            output,
                            sname,
                            i,
                            j,
                            k,
                            center,
                            center_params,
                            edges,
                            edge_params,
                            edge_info,
                            nedges_counted,
                        )
                    )
                    # This rotation guarantees that every competitor is
                    # the center once per scenario per repetition
                    competitors = [competitors[-1]] + competitors[:-1]
        if verbose:
            print(f"Will run {len(jobs)} negotiations")

        def process_info(job: JobInfo, info: SessionInfo):
            center_multiplier = (
                center_multiplier_val
                if center_multiplier_val is not None
                else len(job.edge_info)
            )
            r = info.results
            results.append(info)
            center, center_params = job.center, job.center_params
            cname = (
                type_name(center)
                if not center_params
                else f"{type_name(center)}_{hash(str(center_params))}"
            )
            mean_edge_utility = sum(r.edge_utilities) / len(r.edge_utilities)
            scores.append(
                dict(
                    agent=cname,
                    utility=r.center_utility * center_multiplier,
                    partner_average_utility=mean_edge_utility,
                    scenario=job.sname,
                    repetition=job.rep_index,
                    rotation=job.competitor_index,
                    scenario_index=job.scenario_index,
                    index=0,
                    time=r.total_time,
                    errors=sum(
                        [
                            m.state.has_error and m.state.erred_negotiator == cname
                            for m in r.mechanisms
                        ]
                    ),
                    partner_errors=sum(
                        [
                            m.state.has_error and m.state.erred_negotiator != cname
                            for m in r.mechanisms
                        ]
                    ),
                    # TODO: get the correct number of mechanism errors
                    mechanism_errors=0,
                )
            )
            final_scores[cname] += r.center_utility * center_multiplier
            final_scoresC[cname] += r.center_utility * center_multiplier
            count_center[cname] += 1
            for e, (c, p) in enumerate(job.edge_info[: job.nedges_counted]):
                ename = type_name(c) if not p else f"{type_name(c)}_{hash(str(p))}"
                scores.append(
                    dict(
                        agent=ename,
                        utility=r.edge_utilities[e] * edge_multiplier,
                        partner_average_utility=r.center_utility,
                        scenario=job.sname,
                        repetition=job.rep_index,
                        rotation=job.competitor_index,
                        scenario_index=job.scenario_index,
                        index=e + 1,
                        time=r.times[e],
                        errors=sum(
                            [
                                m.state.has_error and m.state.erred_negotiator == ename
                                for m in r.mechanisms
                            ]
                        ),
                        partner_errors=sum(
                            [
                                m.state.has_error and m.state.erred_negotiator != ename
                                for m in r.mechanisms
                            ]
                        ),
                        # TODO: get the correct number of mechanism errors
                        mechanism_errors=0,
                    )
                )
                final_scores[ename] += r.edge_utilities[e] * edge_multiplier
                final_scoresE[ename] += r.edge_utilities[e] * edge_multiplier
                count_edge[ename] += 1

            if verbose:
                print(f"Center Utility: {r.center_utility}")
                print(f"Edge Utilities: {r.edge_utilities}")
                print(f"Agreement: {r.agreements}")

        if n_jobs is None:
            for job in track(jobs, "Running Negotiations"):
                job, info = run_session(job, dry, verbose)
                process_info(job, info)
        else:
            assert n_jobs > 0
            with ProcessPoolExecutor(max_workers=n_jobs) as executor:
                # Submit all jobs and store the futures
                futures = [
                    executor.submit(run_session, job, dry, verbose) for job in jobs
                ]

                # Process results as they become available
                for future in as_completed(futures):
                    try:
                        job, info = future.result()
                        process_info(job, info)
                    except Exception as e:
                        print(f"Job failed with exception: {e}")

        weighted_average = {}
        for agent in final_scores.keys():
            average_score_E = (
                final_scoresE[agent] / count_edge[agent] if count_edge[agent] > 0 else 0
            )
            average_score_C = (
                final_scoresC[agent] / count_center[agent]
                if count_center[agent] > 0
                else 0
            )
            weighted_average[agent] = 0.5 * (average_score_C + average_score_E)

        return TournamentResults(
            final_scores={k: v for k, v in final_scores.items()},
            edge_count={k: v for k, v in count_edge.items()},
            center_count={k: v for k, v in count_center.items()},
            final_scoresC={k: v for k, v in final_scoresC.items()},
            final_scoresE={k: v for k, v in final_scoresE.items()},
            weighted_average={k: v for k, v in weighted_average.items()},
            scores=scores,
            session_results=results,
            path=path,
        )



from typing import Literal
from random import random, choice
from anl2025.ufun import CenterUFun, SideUFun
from negmas import SAONMI, InverseUFun, PolyAspiration, PresortingInverseUtilityFunction

from negmas.sao.controllers import SAOController, SAOState
from negmas import (
    ConcederTBNegotiator,
    DiscreteCartesianOutcomeSpace,
    ExtendedOutcome,
    LinearTBNegotiator,
    ResponseType,
)
from negmas.outcomes import Outcome
from negmas.sao.negotiators import AspirationNegotiator

ANL2024_AVAILABLE = False
try:
    from anl_agents.anl2024 import Shochan, AgentRenting2024  # type: ignore

    ANL2024_AVAILABLE = True
except Exception:
    pass

__all__ = [
    "ANL2025Negotiator",
    "TimeBased2025",
    "Random2025",
    "Boulware2025",
    "Linear2025",
    "Conceder2025",
    "IndependentBoulware2025",
    "IndependentLinear2025",
    "IndependentConceder2025",
]
if ANL2024_AVAILABLE:
    __all__ += [
        "Shochan2025",
        "AgentRenting2025",
    ]


class ANL2025Negotiator(SAOController):
    """
    Base class of all participant code.

    See the next two examples of how to implement it (`Boulware2025`, `Random2025`).

    Args:
        n_edges: Number of edges for this negotiator. You can access it using self._n_edges
        update_side_ufuns_on_end: Updates the expected outcome for each thread at the end of
                                  each negotiation threads. These expected values are used
                                  when a side-negotiator calculates its side-utility.
        update_side_ufuns_after_receiving_offers: Updates expected outcome for each thread whenever
                                                  an offer is received from a partner
        update_side_ufuns_after_offering: Updates the expected outcome for each thread after it is
                                          offered (i.e. assume the opponent will accept).
        auto_kill: Removes negotiators from the self.negotiators list whenever the negotiation ends.

    Remarks:
        - This class provides some useful members that can be used when developing the negotiation strategy:
            - `self.ufun` : The `CenterUFun` for the center negotiator (for edge negotiators, it will be a normal negmas ufun).
            - `self.negotiators`: Returns a dict mapping negotiator-IDs (for side negotiators) to the corresponding negotiator
                                  and context. You can use the returned negotiator to access the `NMI` (Negotiator-Mechanism-Interface)
                                  for this side negotiator which in turn have access to the state and other services provided by
                                  negmas NMIs. The context has the following members:
               - `ufun`: The side ufun for the thread.
               - `center`: A boolean indicating whether this is a side negotiator for a center or is an edge negotiator.
               - `index`: The thread index for this thread.
            - `self.active_negotiators` / `self.started_negotiators` / `self.to_start_negotiators` / `self.finished_negotiators` / `self.unfinished_negotiators`
              Same as `self.negotiators` but with the corresponding subset of negotiators only.
    """

    def __init__(
        self,
        *args,
        n_edges: int = 0,
        update_side_ufuns_on_end: bool = True,
        update_side_ufuns_after_offering: bool = False,
        update_side_ufuns_after_receiving_offers: bool = False,
        auto_kill: bool = False,
        **kwargs,
    ):
        super().__init__(*args, auto_kill=auto_kill, **kwargs)
        self._n_edges = n_edges
        self._update_side_ufuns_on_end = update_side_ufuns_on_end
        self._update_side_ufuns_after_offering = update_side_ufuns_after_offering
        self._update_side_ufuns_after_receiving_offers = (
            update_side_ufuns_after_receiving_offers
        )

    def init(self):
        """Called after all mechanisms are created to initialize

        Remarks:
            - self.negotiators can be used to access the threads.
            - Each has a negotiator object and a cntxt object.
            - We can pass anything in the cntxt. Currently, we pass the side ufun
            - Examples:
                1. Access the CenterUFun associated with the agent. For edge agents, this will be the single ufun it uses.
                    my_ufun = self.ufun
                2. Access the side ufun associated with each thread. For edge agents this will be the single ufun it uses.
                    my_side_ufuns = [info.context["ufun"] for neg_id, info in self.negotiators.items()]
                    my_side_indices = [info.context["index"] for neg_id, info in self.negotiators.items()]
                    my_side_is_center = [info.context["center"] for neg_id, info in self.negotiators.items()]
                2. Access the side negotiators connected to different negotiation threads
                    my_side_negotiators = [info.negotiator for neg_id, info in self.negotiators.items()]
        """

    def propose(
        self, negotiator_id: str, state: SAOState, dest: str | None = None
    ) -> Outcome | ExtendedOutcome | None:
        """Called to propose an offer for one of the edge negotiators

        Args:
            negotiator_id: The ID of the connection to the edge negotiator.
            state: The state of the negotiation with this edge negotiator.
            dest: The ID of the edge negotiator


        Returns:
            An outcome to offer. In ANL2025, `None` and `ExtendedOutcome` are not allowed
        """
        return super().propose(negotiator_id, state, dest)

    def respond(
        self, negotiator_id: str, state: SAOState, source: str | None = None
    ) -> ResponseType:
        """Called to respond to an offer from an edge negotiator

        Args:
            negotiator_id: The ID of the connection to the edge negotiator.
            state: The state of the negotiation with this edge negotiator.
            dest: The ID of the edge negotiator

        Returns:
            A response (Accept, Reject, or End_Negotiation)

        Remarks:
            - The current offer on the negotiation thread with this edge
              negotiator can be accessed as `state.current_offer`.
        """
        return super().respond(negotiator_id, state, source)

    def set_expected_outcome(self, negotiator_id: str, outcome: Outcome | None) -> None:
        """
        Sets the expected value for a negotiation thread.
        """
        if not isinstance(self.ufun, CenterUFun):
            return
        _, cntxt = self.negotiators[negotiator_id]
        index = cntxt["index"]
        self.ufun.set_expected_outcome(index, outcome)

    def thread_init(self, negotiator_id: str, state: SAOState) -> None:
        """Called when a negotiation thread starts

        Args:
            negotiator_id: Connection ID to this negotiation thread
            state: The state of the negotiation thread at the start of the negotiation.
        """

    def thread_finalize(self, negotiator_id: str, state: SAOState) -> None:
        """Called when a negotiation thread ends

        Args:
            negotiator_id: Connection ID to this negotiation thread
            state: The state of the negotiation thread at the end of the negotiation.
        """

    def on_negotiation_start(self, negotiator_id: str, state: SAOState) -> None:
        super().on_negotiation_start(negotiator_id, state)
        self.thread_init(negotiator_id, state)

    def on_negotiation_end(self, negotiator_id: str, state: SAOState) -> None:
        if self._update_side_ufuns_on_end:
            self.set_expected_outcome(negotiator_id, state.agreement)
        super().on_negotiation_end(negotiator_id, state)
        self.thread_finalize(negotiator_id, state)


class IndependentBoulware2025(ANL2025Negotiator):
    """
    You can participate by an agent that runs any SAO negotiator independently for each thread.
    """

    def __init__(self, **kwargs):
        kwargs["default_negotiator_type"] = AspirationNegotiator
        super().__init__(**kwargs)


class IndependentLinear2025(ANL2025Negotiator):
    """
    You can participate by an agent that runs any SAO negotiator independently for each thread.
    """

    def __init__(self, **kwargs):
        kwargs["default_negotiator_type"] = LinearTBNegotiator
        super().__init__(**kwargs)


class IndependentConceder2025(ANL2025Negotiator):
    """
    You can participate by an agent that runs any SAO negotiator independently for each thread.
    """

    def __init__(self, **kwargs):
        kwargs["default_negotiator_type"] = ConcederTBNegotiator
        super().__init__(**kwargs)


if ANL2024_AVAILABLE:

    class AgentRenting2025(ANL2025Negotiator):
        """
        You can participate by an agent that runs any SAO negotiator independently for each thread.
        """

        def __init__(self, **kwargs):
            kwargs["default_negotiator_type"] = AgentRenting2024  # type: ignore
            super().__init__(**kwargs)

    class Shochan2025(ANL2025Negotiator):
        """
        You can participate by an agent that runs any SAO negotiator independently for each thread.
        """

        def __init__(self, **kwargs):
            kwargs["default_negotiator_type"] = Shochan  # type: ignore
            super().__init__(**kwargs)


class Random2025(ANL2025Negotiator):
    """
    The most general way to implement an agent is to implement propose and respond.
    """

    p_end = 0.000003
    p_reject = 0.99

    def propose(
        self, negotiator_id: str, state: SAOState, dest: str | None = None
    ) -> Outcome | None:
        """Proposes to the given partner (dest) using the side negotiator (negotiator_id).

        Remarks:
        """
        nmi = self.negotiators[negotiator_id].negotiator.nmi
        os: DiscreteCartesianOutcomeSpace = nmi.outcome_space
        return list(os.sample(1))[0]

    def respond(
        self, negotiator_id: str, state: SAOState, source: str | None = None
    ) -> ResponseType:
        """Responds to the given partner (source) using the side negotiator (negotiator_id).

        Remarks:
            - source: is the ID of the partner.
            - the mapping from negotiator_id to source is stable within a negotiation.

        """

        if random() < self.p_end:
            return ResponseType.END_NEGOTIATION

        if (
            random() < self.p_reject
            or float(self.ufun(state.current_offer)) < self.ufun(None)  # type: ignore
        ):
            return ResponseType.REJECT_OFFER
        return ResponseType.ACCEPT_OFFER


class TimeBased2025(ANL2025Negotiator):
    """
    A time-based conceding agent
    """

    def __init__(
        self,
        *args,
        aspiration_type: Literal["boulware"]
        | Literal["conceder"]
        | Literal["linear"]
        | Literal["hardheaded"]
        | float = "boulware",
        deltas: tuple[float, ...] = (1e-3, 1e-1, 2e-1, 4e-1, 8e-1, 1.0),
        reject_exactly_as_reserved: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._curve = PolyAspiration(1.0, aspiration_type)
        self._inverter: dict[str, InverseUFun] = dict()
        self._best: list[Outcome] = None  # type: ignore
        self._mx: float = 1.0
        self._mn: float = 0.0
        self._deltas = deltas
        self._best_margin = 1e-8
        self.reject_exactly_as_reserved = reject_exactly_as_reserved

    def ensure_inverter(self, negotiator_id) -> InverseUFun:
        """Ensures that utility inverter is available"""
        if self._inverter.get(negotiator_id, None) is None:
            _, cntxt = self.negotiators[negotiator_id]
            ufun = cntxt["ufun"]
            inverter = PresortingInverseUtilityFunction(ufun, rational_only=True)
            inverter.init()
            # breakpoint()
            self._mx, self._mn = inverter.max(), inverter.min()
            self._mn = max(self._mn, ufun(None))
            self._best = inverter.some(
                (
                    max(0.0, self._mn, ufun(None), self._mx - self._best_margin),
                    self._mx,
                ),
                normalized=True,
            )
            if not self._best:
                self._best = [inverter.best()]  # type: ignore
            self._inverter[negotiator_id] = inverter

        return self._inverter[negotiator_id]

    def calc_level(self, nmi: SAONMI, state: SAOState, normalized: bool):
        if state.step == 0:
            level = 1.0
        elif (
            # not self.reject_exactly_as_reserved
            # and
            nmi.n_steps is not None and state.step >= nmi.n_steps - 1
        ):
            level = 0.0
        else:
            level = self._curve.utility_at(state.relative_time)
        if not normalized:
            level = level * (self._mx - self._mn) + self._mn
        return level

    def propose(
        self, negotiator_id: str, state: SAOState, dest: str | None = None
    ) -> Outcome | None:
        """Proposes to the given partner (dest) using the side negotiator (negotiator_id).

        Remarks:
        """
        assert self.ufun
        negotiator, cntxt = self.negotiators[negotiator_id]
        inverter = self.ensure_inverter(negotiator_id)
        nmi = negotiator.nmi
        level = self.calc_level(nmi, state, normalized=True)
        ufun: SideUFun = cntxt["ufun"]
        outcome = None
        if self._mx < float(ufun(None)):
            return None
        for d in self._deltas:
            mx = min(1.0, level + d)
            outcome = inverter.one_in((level, mx), normalized=True)
            # print(f"{self.id} found {outcome} at level {(level, mx)}")
            if outcome:
                break
        if not outcome:
            return choice(self._best)
        return outcome

    def respond(
        self, negotiator_id: str, state: SAOState, source: str | None = None
    ) -> ResponseType:
        """Responds to the given partner (source) using the side negotiator (negotiator_id).

        Remarks:
            - source: is the ID of the partner.
            - the mapping from negotiator_id to source is stable within a negotiation.

        """
        assert self.ufun
        _, cntxt = self.negotiators[negotiator_id]
        ufun: SideUFun = cntxt["ufun"]
        nmi = self.negotiators[negotiator_id][0].nmi
        self.ensure_inverter(negotiator_id)
        # end the negotiation if there are no rational outcomes
        level = self.calc_level(nmi, state, normalized=False)

        if self._mx < float(ufun(None)):
            return ResponseType.END_NEGOTIATION

        # print(f"{self.id} got {ufun(state.current_offer)} at level {level}")
        if (self.reject_exactly_as_reserved and level >= ufun(state.current_offer)) or (
            not self.reject_exactly_as_reserved and level > ufun(state.current_offer)
        ):
            return ResponseType.REJECT_OFFER
        return ResponseType.ACCEPT_OFFER

    def thread_finalize(self, negotiator_id: str, state: SAOState) -> None:
        for side in self.negotiators.keys():
            if side == negotiator_id:
                continue
            if side in self._inverter:
                del self._inverter[side]


class Boulware2025(TimeBased2025):
    def __init__(
        self, *args, deltas: tuple[float, ...] = (0.1, 0.2, 0.4, 0.8, 1), **kwargs
    ):
        super().__init__(*args, aspiration_type="boulware", deltas=deltas, **kwargs)


class Linear2025(TimeBased2025):
    def __init__(
        self, *args, deltas: tuple[float, ...] = (0.1, 0.2, 0.4, 0.8, 1), **kwargs
    ):
        super().__init__(*args, aspiration_type="linear", deltas=deltas, **kwargs)


class Conceder2025(TimeBased2025):
    def __init__(
        self, *args, deltas: tuple[float, ...] = (0.1, 0.2, 0.4, 0.8, 1), **kwargs
    ):
        super().__init__(*args, aspiration_type="conceder", deltas=deltas, **kwargs)



from negmas.outcomes import make_os, make_issue, DiscreteCartesianOutcomeSpace
from anl2025.ufun import MaxCenterUFun
from anl2025.scenario import MultidealScenario
from negmas.preferences.generators import generate_ufuns_for

__all__ = ["make_job_hunt_scenario"]


def make_job_hunt_scenario(
    n_employers: int = 4,
    work_days: tuple[int, int] | int = 6,
    salary: tuple[int, int] | list[int] = [100, 150, 200, 250, 300],
    public_graph: bool = True,
    employer_names: list[str] | None = None,
    employee_name: str = "Employee",
):
    if employer_names is None:
        employer_names = [f"employer{_+1:02}" for _ in range(n_employers)]
    os = make_os(
        [make_issue(work_days, name="days"), make_issue(salary, name="salary")],
        name="JobHunt",
    )
    assert isinstance(os, DiscreteCartesianOutcomeSpace)
    ufun_pairs = [
        generate_ufuns_for(os, ufun_names=(f"with_{ename}", f"{ename}"))
        for ename in employer_names
    ]
    side_ufuns = tuple(_[0] for _ in ufun_pairs)
    edge_ufuns = tuple(_[1] for _ in ufun_pairs)
    center_ufun = MaxCenterUFun(
        side_ufuns=side_ufuns,
        n_edges=n_employers,
        outcome_space=os,
        reserved_value=0.0,
        name=employee_name,
    )

    return MultidealScenario(
        center_ufun, edge_ufuns, side_ufuns=side_ufuns, public_graph=public_graph
    )



import pandas as pd
from pathlib import Path


class TargetEvaluator:
    """Evaluates the center utility value of a set of agreements/disagreements"""

    def __init__(self, reserved_value=0.0, values=None):
        self.reserved_value = reserved_value
        if not values:
            values = pd.read_csv(Path(__file__).parent / "center.csv")
            self.values = dict()

            for _, row in values.iterrows():
                self.values[int(row["quantity"])] = float(row["value"])
        else:
            self.values = values
        # read the utility values from the csv vile

    def __call__(self, agreements):
        if not agreements:
            return self.reserved_value

        # outings = dict(zip(self.days, itertools.repeat(0)))
        quantity_sum = 0
        for agreement in agreements:
            if agreement is None:
                continue
            quantity_sum += int(agreement[0])
        return self.values.get(quantity_sum, self.reserved_value)



from anl2025.scenarios.dinners import make_dinners_scenario
from anl2025.ufun import FlatteningCombiner, HierarchicalCombiner
from negmas import DiscreteCartesianOutcomeSpace, make_issue, make_os


def test_flatten_unflatten():
    outcome_spaces = (
        make_os(([make_issue(4), make_issue(5)])),
        make_os(([make_issue(2), make_issue(3), make_issue(5)])),
        make_os(([make_issue(["a", "b"]), make_issue("a", "b")])),
        make_os(([make_issue("a", "b")])),
        make_os(([make_issue(4), make_issue(["zz", "aa"])])),
    )
    assert all(isinstance(_, DiscreteCartesianOutcomeSpace) for _ in outcome_spaces)
    x = ((0, 1), None, ("x", "xyz"), ("34",), (4, "zz"))
    f = (0, 1, None, None, None, "x", "xyz", "34", 4, "zz")
    flattener = FlatteningCombiner(outcome_spaces)
    y = flattener.combined_outcome(x)
    assert y == f, f"{y=}\n{f=}"
    yy = flattener.combined_outcome(x)
    assert y == yy, f"{y=}\n{yy=}"
    yy = flattener.combined_outcome(x)
    assert y == yy, f"{y=}\n{yy=}"
    z = flattener.separated_outcomes(y)
    assert z == x, f"{z=}\n{x=}"
    zz = flattener.separated_outcomes(z)
    assert zz == z, f"{zz=}\n{z=}"


def test_center_ufun_extremes():
    n_days, n_friends = 2, 3
    scenario = make_dinners_scenario(n_friends=n_friends, n_days=n_days)
    center = scenario.center_ufun
    worst, best = center.extreme_outcomes()
    assert center._combiner.combined_outcome(worst) == worst
    assert center._combiner.combined_outcome(best) == best
    mn, mx = center.minmax()
    assert mn == 0
    assert mn < mx
    assert abs(center(worst) - mn) < 1e-3
    assert abs(center(best) - mx) < 1e-3
    assert isinstance(center.outcome_space, DiscreteCartesianOutcomeSpace)
    assert all(
        isinstance(_, DiscreteCartesianOutcomeSpace) for _ in center.outcome_spaces
    )
    if isinstance(center._combiner, FlatteningCombiner):
        assert len(center.outcome_space.issues) == sum(
            len(_.issues)  # type: ignore
            for _ in center.outcome_spaces
        )
    elif isinstance(center._combiner, HierarchicalCombiner):
        assert len(center.outcome_space.issues) == len(center.outcome_spaces)
    assert center.outcome_space.cardinality == pow(n_days + 1, n_friends)
    outcomes = list(center.outcome_space.enumerate())
    for outcome in outcomes:
        edge_outcomes = center._combiner.separated_outcomes(outcome)
        assert edge_outcomes is not None
        assert center.outcome_spaces is not None and len(edge_outcomes) == len(
            center.outcome_spaces
        )
        for edge_outcome, os in zip(edge_outcomes, center.outcome_spaces, strict=True):
            assert edge_outcome is None or edge_outcome in os



import itertools
import pandas as pd
from pathlib import Path


class DinnersEvaluator:
    """Evaluates the center utility value of a set of agreements/disagreements"""

    def __init__(self, reserved_value=0.0, values=None, days=None):
        self.reserved_value = reserved_value
        if values is None:
            # read the utility values from the csv vile
            values = pd.read_csv(Path(__file__).parent / "center.csv")
            self.days = [_ for _ in values.columns if _ != "value"]
            self.values = dict()
            for _, row in values.iterrows():
                self.values[str(tuple(int(row[col]) for col in self.days))] = row[
                    "value"
                ]
        else:
            self.days = days
            self.values = values

    def __call__(self, agreements):
        if not agreements:
            return self.reserved_value
        outings = dict(zip(self.days, itertools.repeat(0)))
        for agreement in agreements:
            if agreement is None:
                continue
            # day is a tuple of one value which is the day selected
            outings[agreement[0]] += 1
        return self.values.get(
            str(tuple(outings[day] for day in self.days)), self.reserved_value
        )



import sys
import copy
from typing import Any, Optional
from negmas.preferences import UtilityFunction
from attr import define, field
from pathlib import Path
from negmas.serialization import serialize, deserialize
from negmas.preferences.generators import generate_multi_issue_ufuns

from negmas.helpers.inout import load, dump
from anl2025.ufun import CenterUFun, SideUFunAdapter
from anl2025.common import (
    TYPES_MAP,
    CENTER_FILE_NAME,
    EDGES_FOLDER_NAME,
    SIDES_FOLDER_NAME,
    TYPE_IDENTIFIER,
    sample_between,
    get_ufun_class,
)


__all__ = ["MultidealScenario", "make_multideal_scenario"]

EPSILON = 1e-6
TRACE_COLS = (
    "time",
    "relative_time",
    "step",
    "negotiator",
    "offer",
    "responses",
    "state",
)


def type_name_adapter(x: str, types_map=TYPES_MAP) -> str:
    if x in types_map:
        return TYPES_MAP[x]
    if x.endswith(("OutcomeSpace", "Issue")) and "." not in x:
        return f"negmas.outcomes.{x}"
    if x.endswith(("CenterUFun",)) and "." not in x:
        return f"anl2025.ufun.{x}"
    if x.endswith(("UtilityFunction", "Fun")) and "." not in x:
        return f"negmas.preferences.{x}"
    return x


@define
class MultidealScenario:
    """Defines the multi-deal scenario by setting utility functions (and implicitly outcome-spaces)"""

    center_ufun: CenterUFun
    edge_ufuns: tuple[UtilityFunction, ...]
    side_ufuns: tuple[UtilityFunction, ...] | None = None
    name: str = ""
    public_graph: bool = True
    code_files: dict[str, str] = field(factory=dict)

    def __attrs_post_init__(self):
        if self.public_graph:
            for e in self.edge_ufuns:
                e.n_edges = self.center_ufun.n_edges  # type: ignore
                e.oucome_spaces = [  # type: ignore
                    copy.deepcopy(_) for _ in self.center_ufun.outcome_spaces
                ]

    def save(
        self,
        base: Path | str,
        as_folder: bool = True,
        python_class_identifier: str = TYPE_IDENTIFIER,
    ):
        base = Path(base)
        name = self.name if self.name else "scenario"
        if as_folder:
            return self.to_folder(
                folder=base / name,
                mkdir=False,
                python_class_identifier=python_class_identifier,
            )
        return self.to_file(
            path=base / f"{name}.yml", python_class_identifier=python_class_identifier
        )

    @classmethod
    def from_folder(
        cls,
        folder: Path | str,
        name: str | None = None,
        public_graph: bool = True,
        python_class_identifier: str = TYPE_IDENTIFIER,
        type_marker=f"{TYPE_IDENTIFIER}:",
    ) -> Optional["MultidealScenario"]:
        """
        Loads a multi-deal scenario from the given folder.

        Args:
            folder: The path to load the scenario from
            name: The name to give to the scenario. If not given, the folder name
            edges_know_details: If given, edge ufuns will have `n_edges`, `outcome_spaces` members
                                that reveal the number of edges in total and the outcome space for each
                                negotiation thread.
            python_class_identifier: the key in the yaml to define a type.
            type_marker: A marker at the beginning of a string to define a type (for future proofing).
        """
        folder = Path(folder)
        folder = folder.resolve()
        center_file = folder / CENTER_FILE_NAME
        if not center_file.is_file():
            return None
        code_files = dict()
        for f in folder.glob("**/*.py"):
            path = str(f.relative_to(folder))
            with open(f, "r") as file:
                code_files[path] = file.read()
        added_paths = []
        if code_files:
            for code_file in code_files.keys():
                module = str(Path(code_file).parent)
                try:
                    del module
                except Exception:
                    pass
            _added_path = str(folder.resolve())
            added_paths.append(_added_path)
            sys.path.insert(0, _added_path)
        dparams = dict(
            python_class_identifier=python_class_identifier,
            type_marker=type_marker,
            type_name_adapter=type_name_adapter,
        )
        center_ufun = deserialize(load(center_file), **dparams)  # type: ignore
        assert isinstance(
            center_ufun, CenterUFun
        ), f"{type(center_ufun)} but expected a CenterUFun \n{center_ufun}"

        def load_ufuns(f: Path | str) -> tuple[UtilityFunction, ...] | None:
            f = Path(f)
            if not f.is_dir():
                return None
            return tuple(
                deserialize(load(_), **dparams)  # type: ignore
                for _ in f.glob("*.yml")
            )

        edge_ufuns = load_ufuns(folder / EDGES_FOLDER_NAME)
        assert edge_ufuns
        for u, os in zip(edge_ufuns, center_ufun.outcome_spaces):
            if u.outcome_space is None:
                u.outcome_space = os
                if isinstance(u, SideUFunAdapter):
                    u._base_ufun.outcome_space = os
        if public_graph:
            for u in edge_ufuns:
                u.n_edges = center_ufun.n_edges  # type: ignore
                u.outcome_spaces = tuple(  # type: ignore
                    copy.deepcopy(_) for _ in center_ufun.outcome_spaces
                )
        side_ufuns = load_ufuns(folder / SIDES_FOLDER_NAME)
        for p in added_paths:
            sys.path.remove(p)

        return cls(
            center_ufun=center_ufun,
            edge_ufuns=tuple(edge_ufuns),
            side_ufuns=side_ufuns,
            name=folder.name if name is None else name,
            code_files=code_files,
        )

    def to_folder(
        self,
        folder: Path | str,
        python_class_identifier: str = TYPE_IDENTIFIER,
        mkdir: bool = False,
    ):
        folder = Path(folder)
        if mkdir:
            name = self.name if self.name else "scenario"
            folder = folder / name
        folder.mkdir(parents=True, exist_ok=True)
        for path, code in self.code_files.items():
            full_path = folder / path
            full_path.parent.mkdir(exist_ok=True, parents=True)
            with open(full_path, "w") as f:
                f.write(code)

        dump(
            serialize(
                self.center_ufun, python_class_identifier=python_class_identifier
            ),
            folder / CENTER_FILE_NAME,
        )

        def save(fname, ufuns):
            if ufuns is None:
                return
            base = folder / fname
            base.mkdir(parents=True, exist_ok=True)
            for u in ufuns:
                dump(
                    serialize(u, python_class_identifier=python_class_identifier),
                    base / f"{u.name}.yml",
                )

        save(EDGES_FOLDER_NAME, self.edge_ufuns)
        save(SIDES_FOLDER_NAME, self.side_ufuns)

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        """Converts the scenario to a dictionary"""
        return dict(
            name=self.name,
            center_ufun=serialize(
                self.center_ufun, python_class_identifier=python_class_identifier
            ),
            edge_ufuns=serialize(
                self.edge_ufuns, python_class_identifier=python_class_identifier
            ),
            side_ufuns=serialize(
                self.side_ufuns, python_class_identifier=python_class_identifier
            ),
        )

    @classmethod
    def from_dict(
        cls, d: dict[str, Any], python_class_identifier=TYPE_IDENTIFIER
    ) -> Optional["MultidealScenario"]:
        return deserialize(d, python_class_identifier=python_class_identifier)  # type: ignore

    def to_file(self, path: Path | str, python_class_identifier=TYPE_IDENTIFIER):
        dump(self.to_dict(python_class_identifier=python_class_identifier), path)

    @classmethod
    def from_file(
        cls, path: Path | str, python_class_identifier=TYPE_IDENTIFIER
    ) -> Optional["MultidealScenario"]:
        return cls.to_dict(load(path), python_class_identifier=python_class_identifier)  # type: ignore


def make_multideal_scenario(
    nedges: int = 5,
    nissues: int = 3,
    nvalues: int = 7,
    # edge ufuns
    center_reserved_value_min: float = 0.0,
    center_reserved_value_max: float = 0.0,
    center_ufun_type: str | type[CenterUFun] = "LinearCombinationCenterUFun",
    center_ufun_params: dict[str, Any] | None = None,
    # edge ufuns
    edge_reserved_value_min: float = 0.1,
    edge_reserved_value_max: float = 0.4,
) -> MultidealScenario:
    ufuns = [generate_multi_issue_ufuns(nissues, nvalues) for _ in range(nedges)]
    edge_ufuns = [_[0] for _ in ufuns]
    for u in edge_ufuns:
        u.reserved_value = sample_between(
            edge_reserved_value_min, edge_reserved_value_max
        )
    # side ufuns are utilities of the center on individual threads (may or may not be used, see next comment)
    side_ufuns = tuple(_[1] for _ in ufuns)
    # create center ufun using side-ufuns if possible and without them otherwise.
    center_r = sample_between(center_reserved_value_min, center_reserved_value_max)
    utype = get_ufun_class(center_ufun_type)
    center_ufun_params = center_ufun_params if center_ufun_params else dict()
    try:
        center_ufun = utype(
            side_ufuns=side_ufuns,
            reserved_value=center_r,
            outcome_spaces=tuple(u.outcome_space for u in side_ufuns),  # type: ignore
            **center_ufun_params,
        )
    except TypeError:
        try:
            center_ufun = utype(
                ufuns=side_ufuns,
                reserved_value=center_r,
                outcome_spaces=tuple(u.outcome_space for u in side_ufuns),  # type: ignore
                **center_ufun_params,
            )
        except TypeError:
            # if the center ufun does not take `ufuns` as an input, do not pass it
            center_ufun = utype(
                reserved_value=center_r,
                outcome_spaces=tuple(u.outcome_space for u in side_ufuns),  # type: ignore
                **center_ufun_params,
            )

    return MultidealScenario(
        center_ufun=center_ufun,
        side_ufuns=side_ufuns,
        edge_ufuns=tuple(edge_ufuns),
    )



import random
from itertools import product
from numpy import argmin
from typing import Iterable
from negmas import LinearAdditiveUtilityFunction, TableFun
from negmas.outcomes import (
    make_os,
    make_issue,
    DiscreteCartesianOutcomeSpace,
    ContiguousIssue,
)
from anl2025.ufun import LambdaCenterUFun
from anl2025.scenario import MultidealScenario


__all__ = ["make_target_quantity_scenario"]

FloatRange = tuple[float, float] | float
IntRange = tuple[int, int] | list[int] | int


def float_in(x: FloatRange):
    if isinstance(x, Iterable):
        return x[0] + (x[1] - x[0]) * random.random()
    return x


def int_in(x: IntRange):
    if isinstance(x, Iterable):
        return random.randint(min(x), max(x))
    return x


class TargetEvaluator:
    """Evaluates the center utility value of a set of agreements/disagreements"""

    def __init__(self, values: dict[int, float], reserved_value=0.0):
        self.reserved_value = reserved_value
        self.values = {int(k): float(v) for k, v in values.items()}

    def __call__(self, agreements):
        if not agreements:
            return self.reserved_value

        # outings = dict(zip(self.days, itertools.repeat(0)))
        quantity_sum = 0
        for agreement in agreements:
            if agreement is None:
                continue
            quantity_sum += int(agreement[0])
        return float(self.values.get(quantity_sum, self.reserved_value))


def make_values(
    qs: list[int], target: int, shortfall: FloatRange, excess: FloatRange
) -> dict[int, float]:
    if target not in qs:
        dists = [abs(_ - target) for _ in qs]
        target = int(argmin(dists))
    else:
        target = qs.index(target)
    if target > max(qs):
        target = max(qs)
    if target < min(qs):
        target = min(qs)
    values = [0.0] * len(qs)
    values[target] = 1.0
    for i in range(target - 1, -1, -1):
        values[i] = max(0.0, values[i + 1] - float_in(shortfall))
    for i in range(target + 1, len(values)):
        values[i] = max(0.0, values[i - 1] - float_in(excess))
    return dict(zip(qs, values))


def make_target_quantity_scenario(
    n_suppliers: IntRange = 4,
    quantity: IntRange = (1, 5),
    target_quantity: IntRange = (2, 20),
    shortfall_penalty: FloatRange = (0.1, 0.3),
    excess_penalty: FloatRange = (0.1, 0.4),
    public_graph: bool = True,
    supplier_names: list[str] | None = None,
    collector_name: str = "Collector",
    collector_reserved_value: FloatRange = 0.0,
    supplier_reserved_values: FloatRange = 0.0,
    supplier_shortfall_penalty: FloatRange | None = (0.3, 0.9),
    supplier_excess_penalty: FloatRange | None = (0.3, 0.9),
) -> MultidealScenario:
    """Creates a target-quantity type scenario

    Args:
        n_suppliers: Number of suppliers (sources/sellers)
        quantity: The range of values for each supplier (if an integer, then the range is (0, quanityt-1))
        target_quantity: The target quantity of the collector (buyer)
        shortfall_penalty: The collector's penalty for buying an item less than target. Can be a range to sample form it
        excess_penalty: The penalty for buying an item more than target. Can be a range to sample form it
        public_graph: Whether edges (suppliers) know n_edges and outcome-spaces of the center (collector)
        supplier_names: Names of suppliers
        collector_name: Name of the collector
        collector_reserved_value: Range or a single value for collector's reserved value
        supplier_reserved_values: Range or a single value for supplier's reserved value
        supplier_shortfall_penalty: suppliers' penalty for buying an item less than their target. Can be a range to sample form it
        supplier_excess_penalty: suppliers' penalty for buying an item more than their target. Can be a range to sample form it

    Remarks:
        - Supplier's target value is sampled uniformly from the range of values
    """
    n_suppliers = int_in(n_suppliers)
    if supplier_shortfall_penalty is None:
        supplier_shortfall_penalty = shortfall_penalty
    if supplier_excess_penalty is None:
        supplier_excess_penalty = excess_penalty
    if supplier_names is None:
        supplier_names = [f"supplier{_+1:02}" for _ in range(n_suppliers)]
    os = make_os([make_issue(quantity, name="quantity")], name="TargetQantity")
    assert isinstance(os, DiscreteCartesianOutcomeSpace)
    quantities = list(os.issues[0].all)
    if isinstance(os.issues[0], ContiguousIssue):
        totals = list(range(n_suppliers * os.issues[0].max_value + 1))
    else:
        totals = sorted(
            list(set([sum(_) for _ in product(*([[0] + quantities] * n_suppliers))]))
        )
    center_ufun = LambdaCenterUFun(
        n_edges=n_suppliers,
        outcome_space=os,
        evaluator=TargetEvaluator(
            values=make_values(
                totals, int_in(target_quantity), shortfall_penalty, excess_penalty
            )
        ),
        name=collector_name,
        reserved_value=float_in(collector_reserved_value),
    )
    edge_ufuns = tuple(
        LinearAdditiveUtilityFunction(
            outcome_space=os,
            reserved_value=float_in(supplier_reserved_values),
            values=(
                TableFun(
                    make_values(
                        quantities,
                        int_in((os.issues[0].min_value, os.issues[0].max_value)),
                        float_in(supplier_shortfall_penalty),
                        float_in(supplier_excess_penalty),
                    )
                ),
            ),
            weights=(1.0,),
            name=name,
        )
        for name in supplier_names
    )

    return MultidealScenario(center_ufun, edge_ufuns, public_graph=public_graph)



from anl2025.scenario import MultidealScenario
from rich.table import Table
import importlib.metadata
from pathlib import Path
from negmas.outcomes.base_issue import unique_name
from negmas.serialization import dump
import pandas as pd
import anl2025
from typing import Annotated
from anl2025.tournament import Tournament
import typer
from rich import print

from anl2025.ufun import CenterUFun
from anl2025.common import TYPE_IDENTIFIER
from anl2025.runner import DEFAULT_METHOD, RunParams, run_generated_session, run_session


app = typer.Typer()
tournament = typer.Typer()
session = typer.Typer()

app.add_typer(
    tournament, name="tournament", help="Creates, manages and runs tournaments"
)

app.add_typer(
    session, name="session", help="Runs single sessions of multi-deal negotiations."
)


def get_package_version(name: str | None = None):
    if not name:
        name = __name__.split(".")[0]
    metadata = importlib.metadata.metadata(name)

    return metadata["Version"]


@app.command(help="Version information")
def version():
    print(f"ANL2025: {get_package_version()}, NegMAS: {get_package_version('negmas')}")


def do_make(
    path: Path | None,
    generated: int | None = None,
    competitor: list[str] = [
        "Boulware2025",
        "Random2025",
        "Linear2025",
        "Conceder2025",
    ],
    nissues: int = 3,
    nvalues: int = 7,
    center_ufun: str = "MaxCenterUFun",
    python_class_identifier: str = TYPE_IDENTIFIER,
    center_reserved_value_min: float = 0.0,
    center_reserved_value_max: float = 0.0,
    nedges: int = 3,
    edge_reserved_value_min: float = 0.1,
    edge_reserved_value_max: float = 0.4,
    nsteps: int = 100,
    keep_order: bool = True,
    atomic: bool = False,
    share_ufuns: bool = True,
    output: Path = Path.home() / "negmas" / "anl2025" / "tournament",
    separate_scenarios: bool = True,
    name: str = "auto",
    method: str = DEFAULT_METHOD,
    public_graph: bool = True,
):
    if generated is None:
        generated = 3 if path is None else 0
    if name == "auto":
        name = unique_name("t", sep="")
    if name:
        output = output / name

    def full_name(x: str) -> str:
        if x in anl2025.negotiator.__all__:
            return f"anl2025.negotiator.{x}"
        return x

    competitors = tuple(full_name(_) for _ in competitor)
    scenarios = []
    if path is not None:
        for f in path.glob("*.yml"):
            try:
                if f.is_file():
                    s = MultidealScenario.from_file(
                        f,
                        python_class_identifier=python_class_identifier,
                    )
                    if not s:
                        continue
                    scenarios.append(s)
            except Exception:
                pass
        for d in path.glob("**/*"):
            try:
                if d.is_dir():
                    s = MultidealScenario.from_folder(
                        d,
                        public_graph=public_graph,
                        python_class_identifier=python_class_identifier,
                    )
                    if not s:
                        continue
                    scenarios.append(s)
            except Exception:
                pass
    leave = False
    if not competitors:
        print("[red]No competitors[/red]!!")
        leave = True
    if not scenarios and not generated:
        print(
            "[red]No scenarios!![/red] No path to existing scenarios given (--scenarios-path) nor are you requesting generating of new paths (--generate)!!"
        )
        leave = True
    if leave:
        return None, None

    t = Tournament.from_scenarios(
        competitors=(competitors),
        scenarios=tuple(scenarios),
        run_params=RunParams(nsteps, keep_order, share_ufuns, atomic, method),
        n_generated=generated,
        nedges=nedges,
        nissues=nissues,
        nvalues=nvalues,
        center_reserved_value_min=center_reserved_value_min,
        center_reserved_value_max=center_reserved_value_max,
        center_ufun_type=center_ufun,
        edge_reserved_value_min=edge_reserved_value_min,
        edge_reserved_value_max=edge_reserved_value_max,
    )
    path = output / "info.yaml"
    t.save(
        path,
        separate_scenarios=separate_scenarios,
        python_class_identifier=python_class_identifier,
    )
    return t, path


def do_run(
    t: Tournament, nreps: int, output: Path, verbose: bool, dry: bool, njobs: int
):
    results = t.run(nreps, output, verbose, dry, n_jobs=njobs if njobs >= 0 else None)
    if len(results.scores) < 1:
        print(
            "No results found!! Make sure that you pass scenarios either using --scenarios or --generate"
        )
        return
    data = pd.DataFrame.from_records(results.scores)
    data["role"] = data["index"].apply(lambda x: "center" if x == 0 else "edge")
    data.to_csv(output / "scores.csv", index=False)
    dump(results.final_scores, output / "final_scores.yaml")
    print(f"Got {len(results.scores)} scores")
    df = data.groupby(["agent", "role"])["utility"].describe().reset_index()
    if len(df) > 0:
        assert isinstance(df, pd.DataFrame)
        print(df_to_table(df, "Score Summary", empty_repeated_values=("agent",)))
    # Create a table for display
    table = Table(title="Final Scores")
    table.add_column("Rank")
    table.add_column("Name", style="blue")
    table.add_column("Score")

    # Add rows to the table

    scores = results.final_scores
    sorted_scores = dict(sorted(scores.items(), key=lambda item: item[1], reverse=True))
    for i, (name, score) in enumerate(sorted_scores.items()):
        table.add_row(str(i + 1), name, f"{score:.3f}")

    # Print the table using rich
    print(table)


def df_to_table(
    df: pd.DataFrame,
    title: str,
    index: bool = False,
    empty_repeated_values: tuple[str, ...] = tuple(),
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.
    Args:
        df (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        title: Title to show at the top
        index: Show or or do not show an index column
    Returns:
        Table: A rich Table instance populated with the DataFrame values.
    """
    table = Table(title=title)
    if index:
        table.add_column("index")

    # Add the columns
    for column in df.columns:
        table.add_column(str(column))

    # Add the rows
    previous_row = ["" for _ in range(len(df.columns))]
    for ind, value_list in enumerate(df.values.tolist()):
        row = [f"{x:.3f}" if isinstance(x, float) else str(x) for x in value_list]
        if empty_repeated_values:
            row = [
                a
                if (c not in empty_repeated_values) or (a != b and isinstance(a, str))
                else ""
                for (a, b, c) in zip(row, previous_row, df.columns)
            ]
        if index:
            table.add_row(str(ind), *row)
        else:
            table.add_row(*row)
        previous_row = row
    return table


@session.command(name="random", help="Runs a multi-deal negotiation session")
def run_random(
    nissues: Annotated[
        int,
        typer.Option(
            help="Number of negotiation issues", rich_help_panel="Outcome Space"
        ),
    ] = 3,
    nvalues: Annotated[
        int,
        typer.Option(
            help="Number of values per negotiation issue",
            rich_help_panel="Outcome Space",
        ),
    ] = 7,
    center: Annotated[
        str,
        typer.Option(help="The type of the center agent", rich_help_panel="Center"),
    ] = "Boulware2025",
    center_ufun: Annotated[
        str,
        typer.Option(
            help="The type of the center ufun: Any ufun defined in anl2025.ufun is OK. Examples are MaxCenterUFun and MeanSMCenterUFUn",
            rich_help_panel="Center",
        ),
    ] = "MaxCenterUFun",
    center_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Minimum value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    center_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Maximim value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    nedges: Annotated[
        int,
        typer.Option(
            help=(
                "Number of Edges (the M side of the 1-M negotiation session). "
                "If you pass this as 0, you can control the edges one by one using --edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = 10,
    edge: Annotated[
        list[str],
        typer.Option(
            help="Types to use for the edges. If nedges is 0, this will define all the edges in order (no randomization)",
            rich_help_panel="Edges",
        ),
    ] = [
        "Boulware2025",
        "Random2025",
        "Linear2025",
        "Conceder2025",
    ],
    edge_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.1,
    edge_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.4,
    nsteps: Annotated[
        int,
        typer.Option(
            help="Number of negotiation steps (see `atomic` for the exact meaning of this).",
            rich_help_panel="Protocol",
        ),
    ] = 100,
    keep_order: Annotated[
        bool,
        typer.Option(
            help="If given, the mechanisms will be advanced in order in every step.",
            rich_help_panel="Protocol",
        ),
    ] = True,
    atomic: Annotated[
        bool,
        typer.Option(
            help=(
                "If given, each step of a mechanism represents a single offer "
                "(from a center or an edge but not both). This may make the logs"
                " wrong though. If --no-atomic (default), a single step corresponds "
                "to one offer form the center and from an edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = False,
    share_ufuns: Annotated[
        bool,
        typer.Option(
            help="Whether or not to share partner ufun up to reserved value. If any ANL2024 agent is used as center or edge, this MUST be True (default)",
            rich_help_panel="Protocol",
        ),
    ] = True,
    dry: Annotated[
        bool,
        typer.Option(
            help="Dry-run. Does not really run anything.",
            rich_help_panel="Output and Logs",
        ),
    ] = False,
    output: Annotated[
        Path,
        typer.Option(
            help="A directory to store the negotiation logs and plots",
            rich_help_panel="Output",
        ),
    ] = Path.home() / "negmas" / "anl2025" / "session",
    name: Annotated[
        str,
        typer.Option(
            help="The name of this session (a random name will be created if not given)",
            rich_help_panel="Output and Logs",
        ),
    ] = "",
    verbose: Annotated[
        bool,
        typer.Option(help="Verbosity", rich_help_panel="Output and Logs"),
    ] = False,
):
    results = run_generated_session(
        center_type=center,
        center_reserved_value_min=center_reserved_value_min,
        center_reserved_value_max=center_reserved_value_max,
        nedges=nedges,
        center_ufun_type=center_ufun,
        edge_reserved_value_min=edge_reserved_value_min,
        edge_reserved_value_max=edge_reserved_value_max,
        edge_types=edge,  # type: ignore
        nissues=nissues,
        nvalues=nvalues,
        nsteps=nsteps,
        verbose=verbose,
        keep_order=keep_order,
        share_ufuns=share_ufuns,
        atomic=atomic,
        output=output,
        name=name,
        dry=dry,
        method=DEFAULT_METHOD,
    )

    cfun = results.center.ufun
    assert isinstance(cfun, CenterUFun)
    side_ufuns = cfun.side_ufuns()
    for i, (e, m, u) in enumerate(
        zip(results.edges, results.mechanisms, side_ufuns, strict=True)  # type: ignore
    ):
        print(
            f"{i:02}: Mechanism {m.name} between ({m.negotiator_ids}) ended in {m.current_step} ({m.relative_time:4.3}) with {m.agreement}: "
            f"Edge Utility = {e.ufun(m.agreement) if e.ufun else 'unknown'}, "
            # f"Side Utility = {u(m.agreement) if u else 'unknown'}"
        )
    print(f"Center Utility: {results.center_utility}")


@session.command(name="run", help="Runs a multi-deal negotiation session")
def run_scenario(
    path: Annotated[
        Path,
        typer.Argument(
            help="Path to a folder to load scenarios from",
            rich_help_panel="Tournament Control",
        ),
    ],
    center: Annotated[
        str,
        typer.Option(help="The type of the center agent", rich_help_panel="Center"),
    ] = "Boulware2025",
    edge: Annotated[
        list[str],
        typer.Option(
            help="Types to use for the edges. If nedges is 0, this will define all the edges in order (no randomization)",
            rich_help_panel="Edges",
        ),
    ] = [
        "Boulware2025",
        "Random2025",
        "Conceder2025",
        "Linear2025",
    ],
    nsteps: Annotated[
        int,
        typer.Option(
            help="Number of negotiation steps (see `atomic` for the exact meaning of this).",
            rich_help_panel="Protocol",
        ),
    ] = 100,
    keep_order: Annotated[
        bool,
        typer.Option(
            help="If given, the mechanisms will be advanced in order in every step.",
            rich_help_panel="Protocol",
        ),
    ] = True,
    atomic: Annotated[
        bool,
        typer.Option(
            help=(
                "If given, each step of a mechanism represents a single offer "
                "(from a center or an edge but not both). This may make the logs"
                " wrong though. If --no-atomic (default), a single step corresponds "
                "to one offer form the center and from an edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = False,
    share_ufuns: Annotated[
        bool,
        typer.Option(
            help="Whether or not to share partner ufun up to reserved value. If any ANL2024 agent is used as center or edge, this MUST be True (default)",
            rich_help_panel="Protocol",
        ),
    ] = True,
    dry: Annotated[
        bool,
        typer.Option(
            help="Dry-run. Does not really run anything.",
            rich_help_panel="Output and Logs",
        ),
    ] = False,
    output: Annotated[
        Path,
        typer.Option(
            help="A directory to store the negotiation logs and plots",
            rich_help_panel="Output",
        ),
    ] = Path.home() / "negmas" / "anl2025" / "session",
    name: Annotated[
        str,
        typer.Option(
            help="The name of this session (a random name will be created if not given)",
            rich_help_panel="Output and Logs",
        ),
    ] = "",
    verbose: Annotated[
        bool,
        typer.Option(help="Verbosity", rich_help_panel="Output and Logs"),
    ] = False,
):
    s = (
        MultidealScenario.from_file(path)
        if path.is_file()
        else MultidealScenario.from_folder(path)
    )
    if s is None:
        print(f"[red]ERROR: [/red] Cannot load scenario from {path}")
        return
    results = run_session(
        scenario=s,
        center_type=center,
        edge_types=edge,  # type: ignore
        nsteps=nsteps,
        verbose=verbose,
        keep_order=keep_order,
        share_ufuns=share_ufuns,
        atomic=atomic,
        output=output,
        name=name,
        dry=dry,
        method=DEFAULT_METHOD,
    )

    cfun = results.center.ufun
    assert isinstance(cfun, CenterUFun)
    side_ufuns = cfun.side_ufuns()
    for i, (e, m, u) in enumerate(
        zip(results.edges, results.mechanisms, side_ufuns, strict=True)  # type: ignore
    ):
        print(
            f"{i:02}: Mechanism {m.name} between ({m.negotiator_ids}) ended in {m.current_step} ({m.relative_time:4.3}) with {m.agreement}: "
            f"Edge Utility = {e.ufun(m.agreement) if e.ufun else 'unknown'}, "
            # f"Side Utility = {u(m.agreement) if u else 'unknown'}"
        )
    print(f"Center Utility: {results.center_utility}")


# @app.command()
# def tournament(ctx: typer.Context):
#     """
#     Manage tournaments.
#     """
#     # This function is called when the "tournament" command is invoked
#     # You can access the subcommand using ctx.invoked_subcommand
#     pass  # Add any logic you want to execute before subcommands


@tournament.command(help="Makes a tournament without running it")
def make(
    scenarios_path: Annotated[
        Path,
        typer.Option(
            help="Path to a folder to load scenarios from",
            rich_help_panel="Tournament Control",
        ),
    ] = None,  # type: ignore
    generate: Annotated[
        int,
        typer.Option(
            help="Number of random scenarios to generate",
            rich_help_panel="Tournament Control",
        ),
    ] = None,  # type: ignore
    competitor: Annotated[
        list[str],
        typer.Option(
            help="Competitor types",
            rich_help_panel="Edges",
        ),
    ] = [
        "Boulware2025",
        "Random2025",
        "Linear2025",
        "Conceder2025",
    ],
    nissues: Annotated[
        int,
        typer.Option(
            help="Number of negotiation issues", rich_help_panel="Outcome Space"
        ),
    ] = 3,
    nvalues: Annotated[
        int,
        typer.Option(
            help="Number of values per negotiation issue",
            rich_help_panel="Outcome Space",
        ),
    ] = 7,
    center_ufun: Annotated[
        str,
        typer.Option(
            help="The type of the center ufun: Any ufun defined in anl2025.ufun is OK. Examples are MaxCenterUFun and MeanSMCenterUFUn",
            rich_help_panel="Center",
        ),
    ] = "MaxCenterUFun",
    python_class_identifier: Annotated[
        str,
        typer.Option(
            help="The identifier identifying types in saved files.",
            rich_help_panel="Tournament Control",
        ),
    ] = TYPE_IDENTIFIER,
    center_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Minimum value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    center_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Maximim value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    nedges: Annotated[
        int,
        typer.Option(
            help=(
                "Number of Edges (the M side of the 1-M negotiation session). "
                "If you pass this as 0, you can control the edges one by one using --edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = 3,
    edge_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.1,
    edge_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.4,
    nsteps: Annotated[
        int,
        typer.Option(
            help="Number of negotiation steps (see `atomic` for the exact meaning of this).",
            rich_help_panel="Protocol",
        ),
    ] = 100,
    keep_order: Annotated[
        bool,
        typer.Option(
            help="If given, the mechanisms will be advanced in order in every step.",
            rich_help_panel="Protocol",
        ),
    ] = True,
    atomic: Annotated[
        bool,
        typer.Option(
            help=(
                "If given, each step of a mechanism represents a single offer "
                "(from a center or an edge but not both). This may make the logs"
                " wrong though. If --no-atomic (default), a single step corresponds "
                "to one offer form the center and from an edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = False,
    share_ufuns: Annotated[
        bool,
        typer.Option(
            help="Whether or not to share partner ufun up to reserved value. If any ANL2024 agent is used as center or edge, this MUST be True (default)",
            rich_help_panel="Protocol",
        ),
    ] = True,
    output: Annotated[
        Path,
        typer.Option(
            help="A directory to store the negotiation logs and plots",
            rich_help_panel="Output",
        ),
    ] = Path.home() / "negmas" / "anl2025" / "tournament",
    separate_scenarios: Annotated[
        bool,
        typer.Option(
            help="Save scenarios in separate folders in the output directory",
            rich_help_panel="Output",
        ),
    ] = True,
    name: Annotated[
        str,
        typer.Option(
            help="The name of this session (a random name will be created if not given)",
            rich_help_panel="Output and Logs",
        ),
    ] = "auto",
    method: Annotated[
        str,
        typer.Option(
            help="The method for conducting the multi-deal negotiation. Supported methods are sequential and ordered",
            rich_help_panel="Mechanism",
        ),
    ] = DEFAULT_METHOD,
):
    if scenarios_path is None and generate is None:
        print(
            "[red]ERROR[/red] You did not specify a scenarios path using --scenarios-path "
            "nor a number of scenarios to generate using --generate. Please specify at least one of them"
        )
    tournament, path = do_make(
        path=scenarios_path,
        generated=generate,
        competitor=competitor,
        nissues=nissues,
        nvalues=nvalues,
        center_ufun=center_ufun,
        python_class_identifier=python_class_identifier,
        center_reserved_value_min=center_reserved_value_min,
        center_reserved_value_max=center_reserved_value_max,
        nedges=nedges,
        edge_reserved_value_min=edge_reserved_value_min,
        edge_reserved_value_max=edge_reserved_value_max,
        nsteps=nsteps,
        keep_order=keep_order,
        atomic=atomic,
        share_ufuns=share_ufuns,
        output=output,
        separate_scenarios=separate_scenarios,
        name=name,
        method=method,
    )
    if tournament:
        print(f"Tournament information is saved in {path}. Use `run` to run it")


@tournament.command(help="Makes and executes a tournament")
def run(
    scenarios_path: Annotated[
        Path,
        typer.Option(
            help="Path to a folder to load scenarios from",
            rich_help_panel="Tournament Control",
        ),
    ] = None,  # type: ignore
    generate: Annotated[
        int,
        typer.Option(
            help="Number of random scenarios to generate",
            rich_help_panel="Tournament Control",
        ),
    ] = None,  # type: ignore
    competitor: Annotated[
        list[str],
        typer.Option(
            help="Competitor types",
            rich_help_panel="Edges",
        ),
    ] = [
        "Boulware2025",
        "Random2025",
        "Conceder2025",
        "Linear2025",
    ],
    nissues: Annotated[
        int,
        typer.Option(
            help="Number of negotiation issues", rich_help_panel="Outcome Space"
        ),
    ] = 3,
    nvalues: Annotated[
        int,
        typer.Option(
            help="Number of values per negotiation issue",
            rich_help_panel="Outcome Space",
        ),
    ] = 7,
    center_ufun: Annotated[
        str,
        typer.Option(
            help="The type of the center ufun: Any ufun defined in anl2025.ufun is OK. Examples are MaxCenterUFun and MeanSMCenterUFUn",
            rich_help_panel="Center",
        ),
    ] = "MaxCenterUFun",
    python_class_identifier: Annotated[
        str,
        typer.Option(
            help="The identifier identifying types in saved files.",
            rich_help_panel="Tournament Control",
        ),
    ] = TYPE_IDENTIFIER,
    center_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Minimum value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    center_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Maximim value for the center reserved value",
            rich_help_panel="Center",
        ),
    ] = 0.0,
    nedges: Annotated[
        int,
        typer.Option(
            help=(
                "Number of Edges (the M side of the 1-M negotiation session). "
                "If you pass this as 0, you can control the edges one by one using --edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = 3,
    edge_reserved_value_min: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.1,
    edge_reserved_value_max: Annotated[
        float,
        typer.Option(
            help="Number of Edges (the M side of the 1-M negotiation session)",
            rich_help_panel="Edges",
        ),
    ] = 0.4,
    nsteps: Annotated[
        int,
        typer.Option(
            help="Number of negotiation steps (see `atomic` for the exact meaning of this).",
            rich_help_panel="Protocol",
        ),
    ] = 100,
    keep_order: Annotated[
        bool,
        typer.Option(
            help="If given, the mechanisms will be advanced in order in every step.",
            rich_help_panel="Protocol",
        ),
    ] = True,
    atomic: Annotated[
        bool,
        typer.Option(
            help=(
                "If given, each step of a mechanism represents a single offer "
                "(from a center or an edge but not both). This may make the logs"
                " wrong though. If --no-atomic (default), a single step corresponds "
                "to one offer form the center and from an edge"
            ),
            rich_help_panel="Protocol",
        ),
    ] = False,
    share_ufuns: Annotated[
        bool,
        typer.Option(
            help="Whether or not to share partner ufun up to reserved value. If any ANL2024 agent is used as center or edge, this MUST be True (default)",
            rich_help_panel="Protocol",
        ),
    ] = True,
    output: Annotated[
        Path,
        typer.Option(
            help="A directory to store the negotiation logs and plots",
            rich_help_panel="Output",
        ),
    ] = Path.home() / "negmas" / "anl2025" / "tournament",
    separate_scenarios: Annotated[
        bool,
        typer.Option(
            help="Save scenarios in separate folders in the output directory",
            rich_help_panel="Output",
        ),
    ] = True,
    name: Annotated[
        str,
        typer.Option(
            help="The name of this session (a random name will be created if not given)",
            rich_help_panel="Output and Logs",
        ),
    ] = "auto",
    method: Annotated[
        str,
        typer.Option(
            help="The method for conducting the multi-deal negotiation. Supported methods are sequential and ordered",
            rich_help_panel="Mechanism",
        ),
    ] = DEFAULT_METHOD,
    dry: Annotated[
        bool,
        typer.Option(
            help="Dry-run. Does not really run anything.",
            rich_help_panel="Output and Logs",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(help="Verbosity", rich_help_panel="Output and Logs"),
    ] = False,
    nreps: Annotated[
        int,
        typer.Option(
            help="Number of random shuffling repetitions",
            rich_help_panel="Tournament Control",
        ),
    ] = 2,
    njobs: Annotated[
        int,
        typer.Option(
            help="Parallelism. -1 for serial, 0 for maximum parallelism, int>0 for specific number of cores and float less than one for a fraction of cores available",
            rich_help_panel="Tournament Control",
        ),
    ] = 1,
):
    if scenarios_path is None and generate is None:
        print(
            "[red]ERROR[/red] You did not specify a scenarios path using --scenarios-path "
            "nor a number of scenarios to generate using --generate. Please specify at least one of them"
        )
    t, path = do_make(
        path=scenarios_path,
        generated=generate,
        competitor=competitor,
        nissues=nissues,
        nvalues=nvalues,
        center_ufun=center_ufun,
        python_class_identifier=python_class_identifier,
        center_reserved_value_min=center_reserved_value_min,
        center_reserved_value_max=center_reserved_value_max,
        nedges=nedges,
        edge_reserved_value_min=edge_reserved_value_min,
        edge_reserved_value_max=edge_reserved_value_max,
        nsteps=nsteps,
        keep_order=keep_order,
        atomic=atomic,
        share_ufuns=share_ufuns,
        output=output,
        separate_scenarios=separate_scenarios,
        name=name,
        method=method,
    )
    if not t or path is None:
        return
    print(f"Tournament information is saved in {path}. Use `run` to run it")
    do_run(t, nreps, path.parent, verbose, dry, njobs)


@tournament.command(help="Executes a tournament made using the make command.")
def execute(
    path: Annotated[
        Path,
        typer.Argument(
            help="Path to the saved yaml file with tournament info",
            rich_help_panel="Input",
        ),
    ],
    nreps: Annotated[
        int,
        typer.Option(
            help="Number of random shuffling repetitions",
            rich_help_panel="Tournament Control",
        ),
    ] = 2,
    dry: Annotated[
        bool,
        typer.Option(
            help="Dry-run. Does not really run anything.",
            rich_help_panel="Output and Logs",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(help="Verbosity", rich_help_panel="Output and Logs"),
    ] = False,
    njobs: Annotated[
        int,
        typer.Option(
            help="Parallelism. -1 for serial, 0 for maximum parallelism, int>0 for specific number of cores and float less than one for a fraction of cores available",
            rich_help_panel="Tournament Control",
        ),
    ] = 1,
    python_class_identifier: Annotated[
        str,
        typer.Option(
            help="The identifier identifying types in saved files.",
            rich_help_panel="Tournament Control",
        ),
    ] = TYPE_IDENTIFIER,
):
    t = Tournament.load(path, python_class_identifier=python_class_identifier)
    do_run(t, nreps, path.parent, verbose, dry, njobs)


if __name__ == "__main__":
    app()



from .common import *  # noqa: F403
from .negotiator import *  # noqa: F403
from .ufun import *  # noqa: F403
from .runner import *  # noqa: F403
from .scenarios import *  # noqa: F403
from .scenario import *  # noqa: F403
from .tournament import *  # noqa: F403

__all__ = (  # type: ignore
    negotiator.__all__  # noqa: F405
    + ufun.__all__  # noqa: F405
    + runner.__all__  # noqa: F405
    + scenario.__all__  # noqa: F405
    + scenarios.__all__  # noqa: F405
    + tournament.__all__  # noqa: F405
)



from abc import ABC, abstractmethod
from copy import deepcopy
from attrs import define, field
from collections import defaultdict
from typing import Any, Iterable, Protocol
from negmas.inout import get_full_type_name
from negmas.outcomes.optional_issue import OptionalIssue
from negmas.serialization import serialize, deserialize
from collections.abc import Sequence, Callable
from enum import Enum
from typing import TypeVar
from negmas.preferences import UtilityFunction, BaseUtilityFunction
from negmas.outcomes import (
    CartesianOutcomeSpace,
    EnumeratingOutcomeSpace,
    Outcome,
    OutcomeSpace,
    make_os,
    make_issue,
)
from negmas.warnings import warn
import numpy as np
from anl2025.common import TYPE_IDENTIFIER

TRACE_COLS = (
    "time",
    "relative_time",
    "step",
    "negotiator",
    "offer",
    "responses",
    "state",
)

__all__ = [
    "CenterUFunCategory",
    "CenterUFun",
    "FlatCenterUFun",
    "LambdaCenterUFun",
    "LambdaUtilityFunction",
    "MaxCenterUFun",
    "LinearCombinationCenterUFun",
    "MeanSMCenterUFun",
    "SideUFun",
    "SingleAgreementSideUFunMixin",
    "UtilityCombiningCenterUFun",
    "FlatteningCombiner",
    "HierarchicalCombiner",
    "DefaultCombiner",
    "OSCombiner",
    "convert_to_center_ufun",
    "make_side_ufun",
]

TUFun = TypeVar("TUFun", bound=UtilityFunction)
CenterEvaluator = Callable[[tuple[Outcome | None, ...] | None], float]
EdgeEvaluator = Callable[[Outcome | None], float]


class CenterUFunCategory(Enum):
    """The type of the center utility function.

    Remarks:
        - `Global`  means that side ufuns are not defined. Note that
          the `side_ufuns` function may still return `SideUfun` objects
          but these will be the global ufun in case all other negotiation
          threads end in disagreement.
        - `Local` means that side ufuns are defined.

    """

    Global = 0
    Local = 1


# def flatten_outcome_spaces(
#     outcome_spaces: tuple[OutcomeSpace, ...],
#     add_index_to_issue_names: bool = False,
#     add_os_to_issue_name: bool = False,
# ) -> CartesianOutcomeSpace:
#     """Generates a single outcome-space which is the Cartesian product of input outcome_spaces."""
#
#     def _name(i: int, os_name: str | None, issue_name: str | None) -> str:
#         x = issue_name if issue_name else ""
#         if add_os_to_issue_name and os_name:
#             x = f"{os_name}:{x}"
#         if add_index_to_issue_names:
#             x = f"{x}:{i+1}"
#         return x
#
#     values, names, nissues = [], [], []
#     for i, os in enumerate(outcome_spaces):
#         if isinstance(os, EnumeratingOutcomeSpace):
#             values.append(list(os.enumerate()))
#             names.append(_name(i, "", os.name))
#             nissues.append(1)
#         elif isinstance(os, CartesianOutcomeSpace):
#             for issue in os.issues:
#                 values.append(issue.values)
#                 names.append(_name(i, os.name, issue.name))
#             nissues.append(len(os.issues))
#         else:
#             raise TypeError(
#                 f"Outcome space of type {type(os)} cannot be combined with other outcome-spaces"
#             )
#     return make_os([make_issue(v, n) for v, n in zip(values, names)])


def unflatten_outcome_space(
    outcome_space: CartesianOutcomeSpace, nissues: tuple[int, ...] | list[int]
) -> tuple[CartesianOutcomeSpace, ...]:
    """Distributes the issues of an outcome-space into a tuple of outcome-spaces."""
    nissues = list(nissues)
    beg = [0] + nissues[:-1]
    end = nissues
    return tuple(
        make_os(outcome_space.issues[i:j], name=f"OS{i}")
        for i, j in zip(beg, end, strict=True)
    )


class OSCombiner(Protocol):
    def __init__(self, outcome_spaces: tuple[OutcomeSpace, ...]) -> None:
        ...

    def combined_space(self) -> CartesianOutcomeSpace:
        ...

    def separated_spaces(self) -> tuple[OutcomeSpace, ...]:
        ...

    def combined_outcome(
        self, outcomes: tuple[Outcome | None, ...] | Outcome | None
    ) -> Outcome | None:
        ...

    def separated_outcomes(
        self, outcome: Outcome | None
    ) -> tuple[Outcome | None, ...] | None:
        ...


def _calc_n_issues(outcome_spaces: tuple[OutcomeSpace, ...]):
    return tuple(
        1 if isinstance(_, EnumeratingOutcomeSpace) else len(_.issues)  # type: ignore
        for _ in outcome_spaces
    )


def is_separated(x: Outcome | tuple[Outcome | None, ...]) -> bool:
    if x is None:
        return True
    if all(_ is None for _ in x):
        return True
    types = set(type(_) for _ in x if _ is not None)
    if all(issubclass(_, tuple) for _ in types):
        return True
    return False


def is_combined(x: Outcome | tuple[Outcome | None, ...]) -> bool:
    if x is None:
        return True
    if any(_ is None for _ in x):
        return False
    types = set(type(_) for _ in x if _ is not None)
    if all(issubclass(_, tuple) for _ in types):
        return False
    return True


@define
class FlatteningCombiner(OSCombiner):
    outcome_spaces: tuple[OutcomeSpace, ...]
    issue_name_format: str = "%osname%:%issuename%"
    combined_name: str = "combined"
    outcome_space: CartesianOutcomeSpace = field(init=False, default=None)

    def __attrs_post_init__(self):
        def _name(i: int, os_name: str | None, issue_name: str | None) -> str:
            x = (
                self.issue_name_format.replace(
                    "%issuename%", issue_name if issue_name is not None else "issue"
                )
                .replace("%osname%", os_name if os_name is not None else "os")
                .replace("%index%", str(i + 1))
            )
            if not x:
                x = "issue"
            return x

        values, names = [], []
        for i, os in enumerate(self.outcome_spaces):
            if isinstance(os, EnumeratingOutcomeSpace):
                values.append(list(os.enumerate()))
                names.append(_name(i, "", os.name))
            elif isinstance(os, CartesianOutcomeSpace):
                for issue in os.issues:
                    values.append(issue.values)
                    names.append(_name(i, os.name, issue.name))
            else:
                raise TypeError(
                    f"Outcome space of type {type(os)} cannot be combined with other outcome-spaces"
                )
        self.outcome_space = make_os(
            [make_issue(v, n) for v, n in zip(values, names)], name=self.combined_name
        )

    def separated_spaces(self) -> tuple[OutcomeSpace, ...]:
        return self.outcome_spaces

    def combined_space(self) -> CartesianOutcomeSpace:
        return self.outcome_space

    def combined_outcome(
        self, outcomes: tuple[Outcome | None, ...] | Outcome | None
    ) -> Outcome | None:
        if not outcomes:
            return outcomes
        if is_combined(outcomes):
            return outcomes
        values = []
        for x, os in zip(outcomes, self.outcome_spaces, strict=True):
            if x is None:
                values += [None] * len(os.issues)  # type: ignore
                continue
            values += list(x)
        return tuple(values)

    def separated_outcomes(
        self, outcome: Outcome | None
    ) -> tuple[Outcome | None, ...] | None:
        if not outcome:
            return outcome
        if is_separated(outcome):
            return outcome
        nxt = 0
        vals = []
        for os in self.outcome_spaces:
            n = len(os.issues)  # type: ignore
            x = outcome[nxt : nxt + n]
            if all(_ is None for _ in x):
                vals.append(None)
            else:
                vals.append(x)
            nxt += n
        return tuple(vals)


@define
class HierarchicalCombiner(OSCombiner):
    outcome_spaces: tuple[OutcomeSpace, ...]
    issue_name_format: str = "agreement%index%"
    combined_name: str = "combined"
    outcome_space: CartesianOutcomeSpace = field(init=False, default=None)

    def __attrs_post_init__(self):
        def _name(i: int, os_name: str | None) -> str:
            x = self.issue_name_format.replace(
                "%osname%", os_name if os_name is not None else "os"
            ).replace("%index%", str(i + 1))
            if not x:
                x = "agreement"
            return x

        values, names = [], []
        for i, os in enumerate(self.outcome_spaces):
            values.append(list(os.enumerate()))  # type: ignore
            names.append(_name(i, os.name))  # type: ignore
        self.outcome_space = make_os(  # type: ignore
            [OptionalIssue(make_issue(v, n), n) for v, n in zip(values, names)],
            name=self.combined_name,
        )

    def separated_spaces(self) -> tuple[OutcomeSpace, ...]:
        return self.outcome_spaces

    def combined_space(self) -> CartesianOutcomeSpace:
        return self.outcome_space

    def combined_outcome(
        self, outcomes: tuple[Outcome | None, ...] | Outcome | None
    ) -> Outcome | None:
        return outcomes

    def separated_outcomes(
        self, outcome: Outcome | None
    ) -> tuple[Outcome | None, ...] | None:
        return outcome


DefaultCombiner = HierarchicalCombiner


def convert_to_center_ufun(
    ufun: UtilityFunction,
    nissues: tuple[int],
    combiner_type: type[OSCombiner] = DefaultCombiner,
    side_evaluators: list[EdgeEvaluator] | None = None,
) -> "CenterUFun":
    """Creates a center ufun from any standard ufun with ufuns side ufuns"""
    assert ufun.outcome_space and isinstance(ufun.outcome_space, CartesianOutcomeSpace)
    evaluator = ufun
    if side_evaluators is not None:
        return LocalEvaluationCenterUFun(
            outcome_spaces=unflatten_outcome_space(ufun.outcome_space, nissues),
            evaluator=evaluator,
            side_evaluators=tuple(side_evaluators),
            combiner_type=combiner_type,
        )
    return LambdaCenterUFun(
        outcome_spaces=unflatten_outcome_space(ufun.outcome_space, nissues),
        evaluator=evaluator,
        combiner_type=combiner_type,
    )


class CenterUFun(UtilityFunction, ABC):
    """
    Base class of center utility functions.

    Remarks:
        - Can be constructed by either passing a single `outcome_space` and `n_edges` or a tuple of `outcome_spaces`
        - It's eval() method  receives a tuple of negotiation results and returns a float
    """

    def __init__(
        self,
        *args,
        outcome_spaces: tuple[CartesianOutcomeSpace, ...] = (),
        n_edges: int = 0,
        combiner_type: type[OSCombiner] = DefaultCombiner,
        expected_outcomes: tuple[Outcome | None, ...]
        | list[Outcome | None]
        | None = None,
        stationary: bool = True,
        stationary_sides: bool | None = None,
        side_ufuns: tuple[BaseUtilityFunction | None, ...] | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not outcome_spaces and self.outcome_space:
            outcome_spaces = tuple([self.outcome_space] * n_edges)
        self._combiner = combiner_type(outcome_spaces)
        self._outcome_spaces = self._combiner.separated_spaces()
        self.n_edges = len(outcome_spaces)
        self.stationary = stationary
        self.stationary_sides = (
            stationary_sides if stationary_sides is not None else stationary
        )
        self.__kwargs = dict(
            reserved_value=self.reserved_value,
            owner=self.owner,
            invalid_value=self._invalid_value,
            name=self.name,
            id=self.id,
        )
        self.__nissues = _calc_n_issues(self._outcome_spaces)
        try:
            self.outcome_space = self._combiner.combined_space()
        except Exception:
            warn("Failed to find the Cartesian product of input outcome spaces")
            self.outcome_space, self.__nissues = None, tuple()
        self._expected: list[Outcome | None] = (
            list(expected_outcomes)
            if expected_outcomes
            else ([None for _ in range(self.n_edges)])
        )

        if side_ufuns is None:
            ufuns = tuple(None for _ in range(self.n_edges))
        else:
            ufuns = tuple(deepcopy(_) for _ in side_ufuns)
        self._effective_side_ufuns = tuple(
            make_side_ufun(self, i, side) for i, side in enumerate(ufuns)
        )

    def is_stationary(self) -> bool:
        return self.stationary

    def set_expected_outcome(self, index: int, outcome: Outcome | None) -> None:
        # print(f"Setting expected outcome for {index} to {outcome}")
        self._expected[index] = outcome
        # print(f"{self._expected}")

    @property
    def outcome_spaces(self) -> tuple[OutcomeSpace, ...]:
        return self._outcome_spaces

    def eval_with_expected(self, offer: Outcome | None, use_expected=True) -> float:
        """Calculates the utility of a set of offers with control over whether or not to use stored expected outcomes."""
        outcomes = self._combiner.separated_outcomes(offer)
        if outcomes:
            if use_expected:
                outcomes = tuple(
                    outcome if outcome else expected
                    for outcome, expected in zip(outcomes, self._expected, strict=True)
                )
            if all(_ is None for _ in outcomes):
                outcomes = None
        if outcomes is None:
            return self.reserved_value
        return self.eval(outcomes)

    def __call__(self, offer: Outcome | None, use_expected=True) -> float:
        """Entry point to calculate the utility of a set of offers (called by the mechanism).

        Override to avoid using expected outcomes."""

        return self.eval_with_expected(offer, use_expected=use_expected)

    @abstractmethod
    def eval(self, offer: tuple[Outcome | None, ...] | Outcome | None) -> float:
        """
        Evaluates the utility of a given set of offers.

        Remarks:
            - Order matters: The order of outcomes in the offer is stable over all calls.
            - A missing offer is represented by `None`
        """

    @abstractmethod
    def ufun_type(self) -> CenterUFunCategory:
        """Returns the center ufun category.

        Currently, we have two categories (Global and Local). See `CenterUFunCategory` for
        their definitions.
        """
        ...

    def side_ufuns(self) -> tuple[BaseUtilityFunction | None, ...]:
        """Should return an independent ufun for each side negotiator of the center."""
        ufuns = self._effective_side_ufuns
        # Make sure that these side ufuns are connected to self
        for i, u in enumerate(ufuns):
            if id(u._center_ufun) == id(self):
                continue
            u._center_ufun = self
            u._index = i
            u._n_edges = self.n_edges
        return ufuns

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        return {
            python_class_identifier: get_full_type_name(type(self)),
            "outcome_spaces": serialize(
                self._outcome_spaces, python_class_identifier=python_class_identifier
            ),
            "name": self.name,
            "reserved_value": self.reserved_value,
        }

    @classmethod
    def from_dict(cls, d, python_class_identifier=TYPE_IDENTIFIER):
        d.pop(python_class_identifier, None)
        for f in ("outcome_spaces", "ufuns"):
            if f in d:
                d[f] = deserialize(
                    d[f], python_class_identifier=python_class_identifier
                )
        return cls(**d)
        # type_ = d.pop(python_class_identifier, cls)
        # # cls = get_class(type_) if isinstance(type_, str) else type_
        # return cls(**d)


class LambdaCenterUFun(CenterUFun):
    """
    A center utility function that implements an arbitrary evaluator
    """

    def __init__(self, *args, evaluator: CenterEvaluator, **kwargs):
        super().__init__(*args, **kwargs)
        self._evaluator = evaluator

    def eval(self, offer: tuple[Outcome | None, ...] | Outcome | None) -> float:
        return self._evaluator(self._combiner.separated_outcomes(offer))

    def ufun_type(self) -> CenterUFunCategory:
        return CenterUFunCategory.Global

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        return super().to_dict(python_class_identifier) | dict(
            evaluator=serialize(
                self._evaluator, python_class_identifier=python_class_identifier
            )
        )


class LambdaUtilityFunction(UtilityFunction):
    """A utility function that implements an arbitrary mapping"""

    def __init__(self, *args, evaluator: EdgeEvaluator, **kwargs):
        super().__init__(*args, **kwargs)
        self._evaluator = evaluator

    def __call__(self, offer: Outcome | None) -> float:
        return self._evaluator(offer)

    def eval(self, offer: Outcome) -> float:
        return self._evaluator(offer)

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        return super().to_dict(python_class_identifier) | dict(
            evaluator=serialize(
                self._evaluator, python_class_identifier=python_class_identifier
            )
        )


class SideUFun(BaseUtilityFunction):
    """
    Side ufun corresponding to the i's component of a center ufun.
    """

    def __init__(
        self,
        *args,
        center_ufun: CenterUFun,
        index: int,
        n_edges: int,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._center_ufun = center_ufun
        self._index = index
        self._n_edges = n_edges

    def is_stationary(self) -> bool:
        return self._center_ufun.stationary_sides

    def set_expected_outcome(
        self, outcome: Outcome | None, index: int | None = None
    ) -> None:
        if index is None:
            index = self._index
        self._center_ufun.set_expected_outcome(index, outcome)

    def eval(self, offer: Outcome | None) -> float:
        exp_outcome = [_ for _ in self._center_ufun._expected]
        offers = [_ for _ in self._center_ufun._expected]
        offers[self._index] = offer

        # For all offers that are after the current agent, set to None:
        for i in range(self._index + 1, self._n_edges):
            # if offers[i] is not None:
            # print(
            #     f"{self._index=}: {i} has a none None outcome ({offers[i]=})\n{offers=}"
            # )
            # raise AssertionError(
            #     f"{self._index=}: {i} has a none None outcome ({offers[i]=})\n{offers=}"
            # )
            self.set_expected_outcome(None, i)
            offers[i] = None
        u = self._center_ufun(tuple(offers))

        # restore expected_outcome:
        for i in range(self._index + 1, self._n_edges):
            self.set_expected_outcome(exp_outcome[i], i)

        return u


def make_side_ufun(
    center: CenterUFun, index: int, side: BaseUtilityFunction | None
) -> SideUFun:
    """Creates a side-ufun for the center at the given index."""
    if side is None:
        side_ufun = SideUFun(center_ufun=center, n_edges=center.n_edges, index=index)
        side_ufun.reserved_value = side_ufun.eval(None)  # type: ignore
    elif isinstance(side, SideUFun):
        side_ufun = side
        side_ufun._center_ufun = center
        side_ufun._index = index
        side_ufun._n_edges = center.n_edges
        side_ufun.reserved_value = side_ufun.eval(None)  # type: ignore
    else:
        side_ufun = SideUFunAdapter(
            center_ufun=center,
            n_edges=center.n_edges,
            index=index,
            base_ufun=side,
            name=side.name,
            id=side.id,
            outcome_space=side.outcome_space,
            reserved_value=side.reserved_value,
            invalid_value=side._invalid_value,
            owner=side.owner,
            type_name=side.type_name,
            reserved_outcome=side.reserved_outcome,
        )
    return side_ufun


class LocalEvaluationCenterUFun(CenterUFun):
    """
    A center utility function that implements an arbitrary evaluator applied to side evaluators
    """

    def __init__(
        self,
        *args,
        evaluator: CenterEvaluator,
        side_evaluators: EdgeEvaluator | Sequence[EdgeEvaluator] | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._evaluator = evaluator
        if not side_evaluators:
            self._side_ufuns = None
            return
        if not isinstance(side_evaluators, Sequence):
            evaluators = [side_evaluators] * len(self._outcome_spaces)
        else:
            evaluators = list(side_evaluators)
        sides: list[LambdaUtilityFunction] = []
        for e, o in zip(evaluators, self._outcome_spaces):
            sides.append(LambdaUtilityFunction(outcome_space=o, evaluator=e))
        self._side_ufuns: tuple[LambdaUtilityFunction, ...] | None = tuple(sides)

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        return super().to_dict(python_class_identifier) | dict(
            side_evaluators=[_._evaluator for _ in self._side_ufuns]
            if self._side_ufuns
            else None,
            evaluator=self._evaluator,
        )

    def eval(self, offer: tuple[Outcome | None, ...] | Outcome | None) -> float:
        return self._evaluator(self._combiner.separated_outcomes(offer))

    def ufun_type(self) -> CenterUFunCategory:
        return CenterUFunCategory.Local


class FlatCenterUFun(UtilityFunction):
    """
    A flattened version of a center ufun.

    A normal CenterUFun takes outcomes as a tuple of outcomes (one for each edge).
    A flattened version of the same ufun takes input as just a single outcome containing
    a concatenation of the outcomes in all edges.

    Example:

        ```python
        x = CenterUFun(...)
        y = x.flatten()

        x(((1, 0.5), (3, true), (7,))) == y((1, 0.5, 3, true, 7))
        ```
    """

    def __init__(
        self, *args, base_ufun: CenterUFun, nissues: tuple[int, ...], **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__base = base_ufun
        self.__nissues = list(nissues)

    def _unflatten(
        self, outcome: Outcome | tuple[Outcome | None, ...] | None
    ) -> tuple[Outcome | None, ...] | None:
        if outcome is None:
            return None

        if isinstance(outcome, tuple) and isinstance(outcome[0], Outcome):
            return outcome

        beg = [0] + self.__nissues[:-1]
        end = self.__nissues
        outcomes = []
        for i, j in zip(beg, end, strict=True):
            x = outcome[i:j]
            if all(_ is None for _ in x):
                outcomes.append(None)
            else:
                outcomes.append(tuple(x))
        return tuple(outcomes)

    def eval(self, offer: Outcome | tuple[Outcome | None] | None) -> float:
        return self.__base.eval(self._unflatten(offer))


class UtilityCombiningCenterUFun(CenterUFun):
    """
    A center ufun with a side-ufun defined for each thread.

    The utility of the center is a function of the ufuns of the edges.
    """

    def __init__(self, *args, side_ufuns: tuple[BaseUtilityFunction, ...], **kwargs):
        super().__init__(*args, **kwargs)
        self.ufuns = tuple(deepcopy(_) for _ in side_ufuns)
        # This is already done in CenterUFun now
        # self._effective_side_ufuns = tuple(
        #     make_side_ufun(self, i, side) for i, side in enumerate(self.ufuns)
        # )

    @abstractmethod
    def combine(self, values: Sequence[float]) -> float:
        """Combines the utilities of all negotiation  threads into a single value"""

    def eval(self, offer: tuple[Outcome | None, ...] | Outcome | None) -> float:
        offer = self._combiner.separated_outcomes(offer)
        if not offer:
            return self.reserved_value
        return self.combine(tuple(float(u(_)) for u, _ in zip(self.ufuns, offer)))

    def ufun_type(self) -> CenterUFunCategory:
        return CenterUFunCategory.Local

    def to_dict(self, python_class_identifier=TYPE_IDENTIFIER) -> dict[str, Any]:
        return super().to_dict(python_class_identifier) | {
            "side_ufuns": serialize(
                self.ufuns, python_class_identifier=python_class_identifier
            ),
            python_class_identifier: get_full_type_name(type(self)),
        }


class MaxCenterUFun(UtilityCombiningCenterUFun):
    """
    The max center ufun.

    The utility of the center is the maximum of the utilities it got in each negotiation (called side utilities)
    """

    def set_expected_outcome(self, index: int, outcome: Outcome | None) -> None:
        # sets the reserved value of all sides
        super().set_expected_outcome(index, outcome)
        r = None
        try:
            set_ufun = self._effective_side_ufuns[index]
        except Exception:
            return

        if isinstance(set_ufun, SideUFunAdapter):
            r = float(set_ufun._base_ufun(outcome))
        elif isinstance(set_ufun, SideUFun):
            return
        if r is None:
            return
        for i, side in enumerate(self._effective_side_ufuns):
            if side is None or i == index:
                continue
            side.reserved_value = max(side.reserved_value, r)
            if isinstance(side, SideUFunAdapter):
                side._base_ufun.reserved_value = max(side._base_ufun.reserved_value, r)

    def combine(self, values: Sequence[float]) -> float:
        return max(values)


class LinearCombinationCenterUFun(UtilityCombiningCenterUFun):
    """
    Linear combination of the side utility values

    The utility of the center is the maximum of the utilities it got in each negotiation (called side utilities)
    """

    def __init__(self, *args, weights: tuple[float, ...] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(weights, Iterable):
            s = sum(weights)
            if s:
                weights = tuple(_ / s for _ in weights)
        self._weights = weights

    def combine(self, values: Sequence[float]) -> float:
        if self._weights is None:
            self._weights = np.random.rand(len(values))
            s = self._weights.sum()
            self._weights = tuple((self._weights / s).tolist())
        return sum(a * b for a, b in zip(values, self._weights, strict=True))


class SideUFunAdapter(SideUFun):
    """Adapts any ufun to be usable as a side-ufun"""

    def __init__(
        self,
        *args,
        base_ufun: BaseUtilityFunction,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._base_ufun = base_ufun

    def to_stationary(self, *args, **kwargs):
        return self._base_ufun.to_stationary(*args, **kwargs)

    def extreme_outcomes(self, *args, **kwargs):
        return self._base_ufun.extreme_outcomes(*args, **kwargs)

    def minmax(self, *args, **kwargs):
        return self._base_ufun.minmax(*args, **kwargs)

    def eval_normalized(self, *args, **kwargs):
        return self._base_ufun.eval_normalized(*args, **kwargs)

    def invert(self, *args, **kwargs):
        return self._base_ufun.invert(*args, **kwargs)

    def scale_by(self, *args, **kwargs):
        return self._base_ufun.scale_by(*args, **kwargs)

    def scale_min_for(self, *args, **kwargs):
        return self._base_ufun.scale_min_for(*args, **kwargs)

    def scale_min(self, *args, **kwargs):
        return self._base_ufun.scale_min(*args, **kwargs)

    def scale_max(self, *args, **kwargs):
        return self._base_ufun.scale_max(*args, **kwargs)

    def normalize_for(self, *args, **kwargs):
        return self._base_ufun.normalize_for(*args, **kwargs)

    def normalize(self, *args, **kwargs):
        return self._base_ufun.normalize(*args, **kwargs)

    def shift_by(self, *args, **kwargs):
        return self._base_ufun.shift_by(*args, **kwargs)

    def shift_min_for(self, *args, **kwargs):
        return self._base_ufun.shift_min_for(*args, **kwargs)

    def shift_max_for(self, *args, **kwargs):
        return self._base_ufun.shift_max_for(*args, **kwargs)

    def argrank_with_weights(self, *args, **kwargs):
        return self._base_ufun.argrank_with_weights(*args, **kwargs)

    def argrank(self, *args, **kwargs):
        return self._base_ufun.argrank(*args, **kwargs)

    def rank_with_weights(self, *args, **kwargs):
        return self._base_ufun.rank_with_weights(*args, **kwargs)

    def rank(self, *args, **kwargs):
        return self._base_ufun.rank(*args, **kwargs)

    def eu(self, *args, **kwargs):
        return self._base_ufun.eu(*args, **kwargs)

    def to_dict(self, *args, **kwargs):
        return self._base_ufun.to_dict(*args, **kwargs)

    def sample_outcome_with_utility(self, *args, **kwargs):
        return self._base_ufun.sample_outcome_with_utility(*args, **kwargs)

    def to_xml_str(self, *args, **kwargs):
        return self._base_ufun.to_xml_str(*args, **kwargs)

    def to_genius(self, *args, **kwargs):
        return self._base_ufun.to_genius(*args, **kwargs)

    def difference_prob(self, *args, **kwargs):
        return self._base_ufun.difference_prob(*args, **kwargs)

    def is_not_worse(self, *args, **kwargs):
        return self._base_ufun.is_not_worse(*args, **kwargs)

    def difference(self, *args, **kwargs):
        return self._base_ufun.difference(*args, **kwargs)

    def max(self):
        return self._base_ufun.max()

    def min(self):
        return self._base_ufun.min()

    def best(self):
        return self._base_ufun.best()

    def worst(self):
        return self._base_ufun.worst()

    def is_volatile(self):
        return self._base_ufun.is_volatile()

    def is_session_dependent(self):
        return self._base_ufun.is_session_dependent()

    def is_state_dependent(self):
        return self._base_ufun.is_state_dependent()

    def scale_max_for(self, *args, **kwargs):
        return self._base_ufun.scale_max_for(*args, **kwargs)

    def to_crisp(self):
        return self._base_ufun.to_crisp()

    def to_prob(self):
        return self._base_ufun.to_prob()

    @property
    def reserved_distribution(self):
        return self._base_ufun.reserved_distribution

    def is_stationary(self):
        return self._base_ufun.is_stationary()

    def changes(self):
        return self._base_ufun.changes()

    def reset_changes(self) -> None:
        return self._base_ufun.reset_changes()

    @property
    def base_type(self) -> str:
        return self._base_ufun.base_type

    def is_better(self, first: Outcome | None, second: Outcome | None) -> bool:
        return self._base_ufun.is_better(first, second)

    def is_equivalent(self, first: Outcome | None, second: Outcome | None) -> bool:
        return self._base_ufun.is_equivalent(first, second)

    def is_not_better(self, first: Outcome | None, second: Outcome | None) -> bool:
        return self._base_ufun.is_not_better(first, second)

    def is_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        return self._base_ufun.is_worse(first, second)

    @property
    def type(self) -> str:
        return self._base_ufun.type


class SingleAgreementSideUFunMixin:
    """Can be mixed with any CenterUFun that is not a combining ufun to create side_ufuns that assume failure on all other negotiations.

    See Also:
        `MeanSMCenterUFun`
    """

    # def side_ufuns(self) -> tuple[BaseUtilityFunction, ...]:
    #     """Should return an independent ufun for each side negotiator of the center"""
    #     return tuple(
    #         SideUFun(center_ufun=self, n_edges=self.n_edges, index=i)  # type: ignore
    #         for i in range(self.n_edges)  # type: ignore
    #     )

    def ufun_type(self) -> CenterUFunCategory:
        return CenterUFunCategory.Local


class MeanSMCenterUFun(SingleAgreementSideUFunMixin, CenterUFun):
    """A ufun that just  returns the average mean+std dev. in each issue of the agreements as the utility value"""

    def eval(self, offer: tuple[Outcome | None, ...] | Outcome | None) -> float:
        offer = self._combiner.separated_outcomes(offer)
        if not offer:
            return 0.0
        n_edges = len(offer)
        if n_edges < 2:
            return 0.1
        vals = defaultdict(lambda: [0.0] * n_edges)
        for e, outcome in enumerate(offer):
            if not outcome:
                continue
            for i, v in enumerate(outcome):
                try:
                    vals[e][i] = float(v[1:])
                except Exception:
                    pass

        return float(sum(np.mean(x) + np.std(x) for x in vals.values())) / len(vals)



from typing import Any
from attr import field
from attrs import define
from random import random
from negmas.helpers.types import get_class

__all__ = [
    "RunParams",
    "TYPE_IDENTIFIER",
    "CENTER_FILE_NAME",
    "EDGES_FOLDER_NAME",
    "SIDES_FOLDER_NAME",
]

EPSILON = 1e-6
TYPE_IDENTIFIER = "type"
CENTER_FILE_NAME = "center.yml"
EDGES_FOLDER_NAME = "edges"
SIDES_FOLDER_NAME = "sides"
TYPES_MAP = dict(
    DiscreteCartesianOutcomeSpace="negmas.outcomes.DiscreteCartesianOutcomeSpace"
)

SEQUENTIAL_METHOD = "sequential"
DEFAULT_METHOD = SEQUENTIAL_METHOD


@define
class RunParams:
    """Defines the running parameters of the multi-deal negotiation like time-limits.

    Attributes:
        nsteps: Number of negotiation steps
        keep_order: Keep the order of  negotiation threads when scheduling next action
        share_ufuns: If given, agents can access partner ufuns through `self.opponeng_ufun`
        atomic: Every step is a single offer (if not given, on the other hand, every step is a complete round)
        method: the method to use for running all the sessions.  Acceptable options are: sequential,
                ordered, threads, processes. See `negmas.mechanisms.Mechanism.run_all()` for full details.
                ANL2025 uses the sequential option which means that a complete negotiation is finished before
                the next starts.
        time_limit: Number of seconds allowed per negotiation
    """

    # mechanism params
    nsteps: int = 100
    keep_order: bool = True
    share_ufuns: bool = False
    atomic: bool = False
    end_on_no_response: bool = True
    method: str = DEFAULT_METHOD
    time_limit: float | None = None
    hidden_time_limit: float | None = None
    mechanism_params: dict[str, Any] = field(factory=dict)
    ignore_mechanism_exceptions: bool = False
    ignore_negotiator_exceptions: bool | None = None


def get_ufun_class(x: str | type) -> type:
    """Returns the type of the agent"""
    if not isinstance(x, str):
        return x
    return get_class(x, module_name="anl2025.ufun")


def get_agent_class(x: str | type) -> type:
    """Returns the type of the agent"""
    if not isinstance(x, str):
        return x
    return get_class(x, module_name="anl2025.negotiator")


def sample_between(mn: float, mx: float, eps: float = EPSILON) -> float:
    """Samples a number if the given range"""
    if (mx - mn) > eps:
        return mn + (mx - mn) * random()
    return (mx + mn) / 2.0
